'''Ugh, work!'''

from concepts.color import *
            
from move import *

import random


class JobBoard(object):
    def __init__(self, city):
        self._city             = city
        self._available_jobs   = set()
        self._unavailable_jobs = set()
        
    def jobs(self):
        return set(self._available_jobs + self._unavailable_jobs)
        
    def available_jobs(self):
        return set(self._available_jobs)
        
    def unavailable_jobs(self):
        return set(self._unavailable_jobs)
    
    def set_available(self, job):
        self._available_jobs.add(job)
        self._unavailable_jobs.discard(job)
        
    def set_unavailable(self, job):
        self._available_jobs.discard(job)
        self._unavailable_jobs.add(job)
    
    def add_job(self, job):
        job.set_on_update(self.update_job)
        
        if job.is_available():
            self.set_available(job)
        else:
            self.set_unavailable(job)
    
    def update_job(self, job):
        self.add_job(job)
        
    def cancel_job(self, job):
        job.cancel()
        self.remove_job(job)
        
    def remove_job(self, job):
        self._available_jobs.discard(job)
        self._unavailable_jobs.discard(job)
        job.set_on_update(None)
        
    def seek_employment(self, person):
        jobs = filter(lambda x: x.building() not in person.prohibited(),
                      self.available_jobs())
        
        if not jobs:
            return
        
        random.choice(list(jobs)).employ(person)
                    
            
class Job(object):
    def __init__(self, building):
        self._building      = building
        self._person        = None
        self._active        = True
        self._must_not_fill = False
        self._callback      = None
    
    def changed(self):
        if self._callback is not None:
            self._callback(self)
    
    def set_on_update(self, callback):
        self._callback = callback
        
    def must_not_fill(self):
        return self._must_not_fill
    
    def allow_filling(self):
        self._must_not_fill = False
        self.changed()
        
    def forbid_filling(self):
        self._must_not_fill = True
        self.fire_worker()
        self.changed()
        
    def title(self):
        return 'Worker'
    
    def color(self):
        return Color(255, 0, 0)
    
    def body_weight(self):
        return 0
    
    def mind_weight(self):
        return 0
    
    def spirit_weight(self):
        return 0
    
    def total_weights(self):
        return self.body_weight() + self.mind_weight() + self.spirit_weight()
    
    def skill(self):
        if self.person() is None:
            return 0.0
        
        weights = self.total_weights()
        total   = (self.person().body() * self.body_weight() +
                   self.person().mind() * self.mind_weight() +
                   self.person().spirit() * self.spirit_weight())
        
        return float(total) / (25.0 * weights)
    
    def performance(self):
        if self.person() is None:
            return 0.0
        
        return min(1.0, self.person().happiness().modifier() * self.skill())
    
    def create_goods(self, type, min, max):
        num = min
        
        for _ in range(max - min):
            if random.random() < self.performance():
                num += 1

        if num <= 0:
            return None
        
        return type(num)
    
    def spawn(self, type, min, max):
        if self.person() is None:
            return
        
        resource = self.create_goods(type, min, max)
        
        if resource is None:
            return
        
        self.person().tile().spawn(resource)
   
    def building(self):
        return self._building
    
    def person(self):
        return self._person
    
    def plan(self):
        return GoAndDoJob(self.person(), self)
    
    def has_plan(self):
        return True
    
    def office(self):
        return self.building().office()
    
    def target(self):
        return self.building().target()
        
    def work_done(self):
        pass
    
    def acted(self):
        pass
    
    def work_speed(self):
        return 0.02
    
    def employ(self, person):
        if self.person() == person:
            return
        
        self.fire_worker()
        self._person = person
        
        if person is not None:
            self.person().employed(self)
            self.changed()
            
    def fire_worker(self):
        if self.person() is None:
            return
        
        self.person().fired(self)
        self._person = None
        self.changed()
    
    def cancel(self):
        self.fire_worker()
        self._active = False
        
    def is_active(self):
        return self._active
    
    def is_available(self):
        return (self._person is None and 
                self.is_active() and
                not self.must_not_fill())

    def __str__(self):
        return self.__class__.__name__
    
    
class Production(Action):
    def __init__(self, job):
        super(Production, self).__init__(job.person())
        
        self._job      = job
        self._progress = 0.0
    
    def speed(self):
        return self.job().work_speed()
    
    def job(self):
        return self._job
    
    def progress(self):
        return self._progress
    
    def unit_produced(self):
        self.job().work_done()
    
    def act(self):
        if self.job().person() != self.person():
            self.done()
        
        self._progress += self.speed()
        self.job().acted()
        
        if self.progress() >= 1.0:
            self.unit_produced()
            self.done()

    
class GoAndDoJob(Plan):
    def __init__(self, person, job, resource = None):
        super(GoAndDoJob, self).__init__(person)
        
        self._job     = job
        self._route   = None
        self._did_job = False
        self._target  = job.target()
        
    def color(self):
        return Color(0, 0, 255)
        
    def job(self):
        return self._job
    
    def target(self):
        return self._target
    
    def route(self):
        return self._route
    
    def did_job(self):
        return self._did_job
    
    def next_action(self):
        if self.job().person() != self.person():
            return None
        
        if self.route() is not None:
            action = self.route().next_action()
            
            if action is not None:
                return action
            
        if self.did_job():
            return None
        elif self.target() is None:
            if self.person().is_at(self.job().office()):
                return None
            else:
                self._route = self.person().route_to(self.job().office())
        elif not self.person().is_at(self.target()):
            self._route = self.person().route_to(self.target())
        elif random.random() < .05:
            self._route = self.person().route_to(self.job().office())
        else:
            self._did_job = True
            return Production(self.job())
        
        return self.next_action()
                
    def path(self):
        if self.route() is None:
            return []
        
        return self.route().path()


class FetchResource(Plan):
    def __init__(self, person, target, resource, predicate = None,
                 then = None):
        super(FetchResource, self).__init__(person)
        
        if predicate is None:
            predicate = lambda g, t: g.matches(self.resource())
        
        if then is None:
            then = person.drop

        self._resource  = resource
        self._target    = target
        self._route     = None
        self._predicate = predicate
        self._then      = then
        
    def color(self):
        return Color(128, 255, 128)
        
    def resource(self):
        return self._resource
    
    def target(self):
        return self._target
    
    def predicate(self):
        return self._predicate
    
    def route(self):
        return self._route
    
    def then(self):
        return self._then
    
    def next_action(self):
        if self._route is not None:
            action = self.route().next_action()
            
            if action is not None:
                return action

                tile = p.find_resource(needed)
                
                if tile is None:
                    return None
                
        backpack = self.person().backpack()
        
        if backpack is not None:
            if self.person().is_at(self.target()):
                self.then()()                    
                return None
            else:
                self._route = self.person().route_to(self.target())
        else:
            location = self.person().tile()
            
            if location.has_goods() and self.predicate()(location.goods(),
                                                         location):
                self.person().take(self.resource())
            else:
                tile = self.person().find_matching_resource_tile(self.predicate())
                
                if tile is None:
                    return None
                
                self._route = self.person().route_to(tile)
    
        return self.next_action()

    def path(self):
        if self.route() is None:
            return []
        
        return self.route().path()
