'''Buildings in River.'''

from concepts.color import *
from concepts.geometry import *

from goods import *

from river.actions.jobs import *

from terrain import *

import random
import operator


class Building(object):
    _types = {'living'    : [None, None, None, None],
              'processing': [None, None, None, None],
              'infra'     : [None, None, None, None],
              'industry'  : [None, None, None, None]}
    
    def __init__(self, city, tile, jobs = []):
        self._city    = city
        self._tile    = tile
        self._jobs    = jobs
        self._office  = random.choice(list(self.tiles()))
        self._targets = filter(lambda x: x != self.office(),
                               list(self.tiles()))
        
    @classmethod
    def categories(cls):
        return ['infra', 'living', 'industry', 'processing']
    
    @classmethod
    def max_buildable(self):
        return None
    
    @classmethod
    def too_many(self):
        return None
    
    @classmethod
    def types(cls, category):
        return list(cls._types[category])
    
    @classmethod
    def add_type(cls, type, category, nr):
        cls._types[category][nr - 1] = type
        
    @classmethod
    def can_be_built_on(cls, terrain):
        return (terrain.__class__ in cls.built_on())
    
    @classmethod
    def stands_free(cls):
        return True

    @classmethod
    def built_on(cls):
        return [Grass]
    
    @classmethod
    def size(self):
        return 1
    
    @classmethod
    def color(self):
        return Color(255, 0, 0)
    
    @classmethod
    def difficulty(self):
        return 1
    
    @classmethod
    def name(cls):
        return 'Building'
    
    @classmethod
    def cost(self):
        return [Wood(3)]
    
    def people(self):
        return reduce(operator.add, map(self.city().persons_on, self.tiles()))

    def destroy(self):
        self.city().destroy(self)
        self._tile = None
        self._jobs = None
        
    def center(self):
        return self._tile.vector() + Vector(self.size() / 2.0,
                                            self.size() / 2.0)
    
    def city(self):
        return self._city
    
    def tiles(self):
        if self._tile is None:
            return []
        
        return self._tile.tiles_lower_right(self.size())

    def random_tile(self):
        if self.size() == 1:
            return self._tile
        
        return random.choice(self._targets)
    
    def area(self):
        if self._tile is None:
            return None
        
        return Area(self._tile.vector(), Vector(self.size(), self.size()))
    
    def is_at(self, tile):
        return (tile in self.tiles())
    
    def jobs(self):
        if self._jobs is None:
            return []
        
        return list(self._jobs)
    
    def housing_space(self):
        return 0
    
    def office(self):
        return self._office
    
    def target(self):
        if self.size() == 1:
            return self._tile
        
        possible = filter(lambda x: not x.is_full(), self._targets)
        
        if not possible:
            return None
        
        self._targets.remove(possible[0])
        self._targets.append(possible[0])
        
        return possible[0]
    
    def resources(self):
        for x in self.tiles():
            if x.has_goods():
                yield x.goods()
    
    def resources_matching(self, predicate):
        return filter(predicate, self.resources())
    
    def tile_for_resource(self, resource):
        for x in self.tiles():
            if x.goods() is resource:
                return x
            
        return None
    
    def __str__(self):
        return '%s at %s' % (self.__class__.__name__, self.area())


class ConstructionSite(Building):
    def __init__(self, city, tile, type):
        self._type = type

        super(ConstructionSite, self).__init__(city, tile,
                                               [Construction(self)])
    
        self._progress       = 0.0
        self._cost_remaining = self.type().cost()
        self._color          = self.type().color()
        self._total          = 0
        self._successor      = None
        
        for x in self._cost_remaining:
            self._total += x.number()
        
    @classmethod
    def difficulty(cls):
        return None
    
    @classmethod
    def name(cls):
        return 'Construction Site'

    @classmethod
    def built_on(cls):
        return None
        
    @classmethod
    def size(cls):
        return None

    def successor(self):
        return self._successor
    
    def goods_needed(self):
        return list(self._cost_remaining)
    
    def reduce_cost(self):
        r = self._cost_remaining[0]
        r.decrease(1)
        
        if r.number() <= 0:
            self._cost_remaining.pop(0)
            
    def total_needed(self):
        return self._total
    
    def number_needed(self):
        total = 0
        
        for x in self._cost_remaining:
            total += x.number()
        
        return total
    
    def type(self):
        return self._type
    
    def size(self):
        return self.type().size()
    
    def progress(self):
        return self._progress
    
    def needs_next_resource(self):
        max = 1 - float(self.number_needed()) / float(self.total_needed())
        
        return (self.progress() >= max)
    
    def next_resource(self):
        if not self._cost_remaining:
            return None
        
        return self._cost_remaining[0]
            
    def add_progress(self, x):
        self._progress += x
        
        max = 1 - float(self.number_needed()) / float(self.total_needed())
            
        if self.progress() > max:
            self._progress = max
            
        if self.progress() >= 1:
            self._progress = 1.0
            tile = self._tile
            type = self.type()
            
            self.city().log('%s constructed.' % (self.type().name()))
        
            self.color = lambda: type.color()            
            self.destroy()
            
            self._successor = self.city().build(tile, type)
            
    def color(self):
        return Color(96, 96, 96).mix(self._color, 0.1 + self.progress() * 0.6)


class Construction(Job):
    def __init__(self, site):
        super(Construction, self).__init__(site)
        
    def title(self):
        return 'Builder'
    
    def color(self):
        return Color(64, 64, 255)
    
    def body_weight(self):
        return 2
    
    def mind_weight(self):
        return 1
    
    def spirit_weight(self):
        return 0
    
    def work_done(self):
        if self.person() is None:
            return
        
        self.building().add_progress(0.01 + self.performance() /
                                     self.building().total_needed())
    
    def work_speed(self):
        return min(1.0, 2.0 / self.building().total_needed())

    def route_to_office(self):
        return self.person().route_to(self.office())

    def route_to_target(self):
        return self.person().route_to(self.target())

    def fetch_resource(self, resource):
        return FetchResource(self.person(), self.target(), resource)
    
    def do_work(self):
        return super(Construction, self).plan()
    
    def plan(self):
        b = self.building()
        p = self.person()
        
        if p is None:
            return
        
        if b.needs_next_resource():
            needed = b.next_resource().with_number(1)
            
            if p.tile().has_goods() and p.tile().goods().is_at_least(needed):
                p.tile().goods().decrease(1)
                b.reduce_cost()
            else:
                return self.fetch_resource(b.next_resource())

        if p.tile() == b.office() and b.size() > 1:
            return self.route_to_target()
            
        return self.do_work()


class ProcessingBuilding(Building):
    def __init__(self, city, tile, jobs, resources):
        super(ProcessingBuilding, self).__init__(city, tile, jobs)
    
        self._needed        = None
        self._all_resources = list(resources)
        self._total         = reduce(lambda a, b: a + b,
                                     map(lambda x: x.number(), resources))
    
        self.restart()
        
    def total(self):
        return self._total
    
    def all_resources(self):
        return list(self._all_resources)
    
    def progress(self):
        return 1.0 - (float(self.total_needed()) / self.total())
    
    def progress_of(self, resource):
        return 1.0 - (float(self.needed_of(resource)) /
                      self.number_of(resource))
        
    def needed(self):
        return list(self._needed)
    
    def needed_of(self, resource):
        for x in self.needed():
            if x.matches(resource):
                return x.number()
            
        return 0
    
    def number_of(self, resource):
        for x in self.all_resources():
            if x.matches(resource):
                return x.number()
            
        return 0
    
    def next_needed(self):
        return self.needed()[0]
    
    def clean_needed(self):
        if self._needed[0].number() <= 0:
            del self._needed[0]
            
    def total_needed(self):
        return reduce(lambda a, b: a + b,
                      map(lambda x: x.number(), self.needed()), 0)
        
    def restart(self):
        self._needed = map(lambda x: x.copy(), self.all_resources())
        
    def is_full(self):
        return (self.total_needed() <= 0)
    
    def resource_added(self, x):
        pass
    
    def add_resource(self, x):
        if x is None:
            return
        
        next = self.next_needed()
        
        if not x.matches(next) or self.is_full():
            return
        
        take = min(x.number(), self.next_needed().number())

        next.decrease(take)
        x.decrease(take)
        
        self.resource_added(x.with_number(take))
        self.clean_needed()


class ProcessingJob(Job):
    def __init__(self, building):
        super(ProcessingJob, self).__init__(building)
    
    def work_completly_done(self):
        pass
    
    def work_done(self):
        if self.person() is None:
            return

        if self.building().is_full():
            self.work_completly_done()
            self.building().restart()

    def route_to_office(self):
        return self.person().route_to(self.office())

    def route_to_target(self):
        return self.person().route_to(self.target())

    def fetch_resource(self, resource):
        return FetchResource(self.person(), self.target(), resource,
                             then = self.handle_raw_material)
    
    def do_work(self):
        return super(ProcessingJob, self).plan()
    
    def handle_raw_material(self):
        self.building().add_resource(self.person().backpack())            
    
    def base_work_speed(self):
        return 0.01
    
    def work_speed(self):
        return self.base_work_speed() / (2 - self.performance())
    
    def plan(self):
        b = self.building()
        p = self.person()
        
        if p is None:
            return
        
        if p.is_at(b.office()) and b.size() > 1:
            return self.route_to_target()
            
        if b.is_full():
            return self.do_work()
        
        return self.fetch_resource(b.next_needed())
            

class Room(object):
    def __init__(self, building):
        self._building      = building
        self._occupant      = None
        self._must_not_fill = False
        
    def building(self):
        return self._building
    
    def occupant(self):
        return self._occupant
    
    def set_occupant(self, occupant):
        self._occupant = occupant
        
    def must_not_fill(self):
        return self._must_not_fill
    
    def evict(self):
        if self.occupant() is None:
            return
        
        self._occupant.evicted(self.building())
        self._occupant = None
    
    def forbid_filling(self):
        self.evict()
        self._must_not_fill = True
    
    def allow_filling(self):
        self._must_not_fill = False
    
    def is_available(self):
        return (self.occupant() is None and not self.must_not_fill())
    
    def space(self):
        if self.is_available():
            return 1
        else:
            return 0
        
        
class House(Building):
    def __init__(self, city, tile):
        super(House, self).__init__(city, tile)
    
        self._rooms = map(lambda x: Room(self), range(2))
        
    @classmethod
    def name(cls):
        return 'House'
    
    @classmethod
    def size(self):
        return 2
    
    @classmethod
    def color(self):
        return Color(128, 0, 0)
    
    @classmethod
    def cost(self):
        return [Wood(4)]
    
    def office(self):
        return None
    
    def room(self, nr):
        return self._rooms[nr]
    
    def rooms(self):
        return list(self._rooms)
    
    def inhabitants(self):
        return set(filter(lambda x: x is not None,
                          map(lambda x: x.occupant(), self.rooms())))
    
    def evict(self, person):
        for x in self.rooms():
            if x.occupant() == person:
                x.evict()
                
    def housing_space(self):
        return reduce(lambda nr, room: nr + room.space(), self.rooms(), 0)

    def move_in(self, person):
        for x in self.rooms():
            if x.is_available():
                x.set_occupant(person)
                return
            
    def move_out(self, person):
        for x in self.rooms():
            if x.occupant() == person:
                x.set_occupant(None)

Building.add_type(House, 'living', 1)
    

class Road(Building):
    def __init__(self, city, tile):
        super(Road, self).__init__(city, tile)
    
    @classmethod
    def name(cls):
        return 'Road'
    
    @classmethod
    def stands_free(cls):
        return False
    
    @classmethod
    def size(self):
        return 1
    
    @classmethod
    def cost(self):
        return [Stone(1)]
    
    @classmethod
    def built_on(cls):
        return [Grass, Forest, Rock, MarbleRock, Mountain]
    
    @classmethod
    def color(self):
        return Color(156, 156, 156)
    
    @classmethod
    def difficulty(self):
        return 0.5
    
    def office(self):
        return None


Building.add_type(Road, 'infra', 2)
    

class Bridge(Building):
    def __init__(self, city, tile):
        super(Bridge, self).__init__(city, tile)
    
    @classmethod
    def name(cls):
        return 'Bridge'
    
    @classmethod
    def stands_free(cls):
        return False
    
    @classmethod
    def built_on(cls):
        return [River]
    
    @classmethod
    def size(self):
        return 1
    
    @classmethod
    def cost(self):
        return [Wood(5)]
    
    @classmethod
    def color(self):
        return Color(72, 64, 0)
    
    @classmethod
    def difficulty(self):
        return 0.6
    
    def office(self):
        return None


Building.add_type(Bridge, 'infra', 1)
    