'''The city in River.'''

from building import *
from map import *
from people import *

from actions.jobs import *

from industries.mythology import *


class City(object):
    def __init__(self, name, map, time):
        self._name      = name
        self._map       = map
        self._time      = time
        self._buildings = set()
        self._persons   = set()
        self._goods     = set()
        self._jobs      = JobBoard(self)
        self._logger    = None
        self._god       = God(self)
        
        map.set_city(self)
    
    def god(self):
        return self._god
    
    def logger(self):
        return self._logger
    
    def set_logger(self, logger):
        self._logger = logger
        
    def has_logger(self):
        return (self._logger is not None)
    
    def log(self, string, color = None):
        if self.has_logger():
            self.logger().log(string, color)
            
    def name(self):
        return self._name
    
    def map(self):
        return self._map
    
    def time(self):
        return self._time
    
    def jobs(self):
        return self._jobs
    
    def buildings(self):
        return set(self._buildings)
    
    def building_on(self, tile):
        result = filter(lambda x: x.is_at(tile), self.buildings())
        
        if not result:
            return None
        
        return result[0]
    
    def has_name(self, name):
        if name is None:
            return True
        
        return any(map(lambda x: x.name() == name, self.persons()))
    
    def persons(self):
        return set(self._persons)
    
    def persons_on(self, tile):
        return filter(lambda x: x.is_at(tile), self.persons())        
        
    def persons_with(self, predicate = lambda x: True):
        return filter(predicate, self.persons())
    
    def persons_of_type(self, type):
        return self.persons_with(lambda x: x.job_description() == type)
    
    def change_attributes(self, delta):
        for x in self.persons():
            x.attributes().change(delta)
    
    def change_happiness(self, delta):
        for x in self.persons():
            x.happiness().add(delta)
    
    def change_nutrition(self, delta):
        for x in self.persons():
            x.nutrition().add(delta)
    
    def change_social(self, delta):
        for x in self.persons():
            x.social().add(delta)
        
    def goods(self):
        return set(self._goods)
    
    def goods_at(self, place):
        return filter(lambda x: x.is_at(place), self.goods())
    
    def resources_of_type(self, x):
        return ResourceOverview(self.goods()).number_of(x)
    
    def destroy_resource(self, type):
        for x in self.goods():
            if x.matches(type(1)):
                x.destroy()
    
    def can_build(self, tile, building_type):
        tiles = set(tile.tiles_lower_right(building_type.size()))
        built = any(map(lambda x: x.is_built(), tiles))
        terrain_matches = all(map(lambda x: building_type.can_be_built_on(x.terrain()), tiles))
                                  
        return (not built and terrain_matches)
    
    def build(self, tile, building_type):
        if not self.can_build(tile, building_type):
            return
        
        if building_type.max_buildable() is not None:
            if len(self.buildings_of_type(building_type)) >= building_type.max_buildable():
                self.log(building_type.too_many())
                return
            
        building = building_type(self, tile)
        
        self._buildings.add(building)
        
        for tile in building.tiles():
            tile.set_building(building)
        
        for job in building.jobs():
            self.jobs().add_job(job)
            
        return building
    
    def start_construction(self, tile, building_type):
        if not self.can_build(tile, building_type):
            return
        
        site = ConstructionSite(self, tile, building_type)
        
        self._buildings.add(site)
        
        for tile in site.tiles():
            tile.set_building(site)
        
        for job in site.jobs():
            self.jobs().add_job(job)
    
    def destroy(self, building):
        self._buildings.discard(building)
        
        for tile in building.tiles():
            tile.set_building(None)
        
        for job in building.jobs():
            self.jobs().cancel_job(job)
        
    def give_birth(self, tile, name = None):
        self._persons.add(Person(name, self, tile))
    
    def add_good(self, good):
        self._goods.add(good)        
        
    def remove_good(self, good):
        self._goods.remove(good)
    
    def tiles(self):
        return self.map().tiles()
    
    def tiles_with(self, predicate):
        return filter(predicate, self.tiles())
    
    def nearest_tile_with(self, person, predicate):
        tiles = self.tiles_with(predicate)
        
        if not tiles:
            return None
        
        position = person.tile()
        tiles.sort(lambda x, y: int(position.distance(x) - 
                                    position.distance(y)))
        
        return tiles[0]
        
    def nearest_tile_with_good(self, person, predicate = lambda g, t: True):
        return self.nearest_tile_with(person, lambda x: x.has_goods()
                                      and predicate(x.goods(), x))
        
    def buildings_with(self, predicate = lambda x: True):
        return filter(predicate, self.buildings())
    
    def buildings_of_type(self, type):
        return self.buildings_with(lambda x: x.__class__ is type)
    
    def nearest_building(self, person, predicate = lambda x: True):
        buildings = filter(predicate, self.buildings())
        
        if not buildings:
            return None
        
        buildings.sort(lambda x, y: int(person.tile().distance(x.center()) - 
                                        person.tile().distance(y.center())))
        
        return buildings[0]
    
    def nearest_building_of_type(self, person, type):
        return self.nearest_building(person, lambda x: x.__class__ is type)
    
    def seek_home(self, person):
        home = self.nearest_building(person, lambda x: x.housing_space() > 0
                                     and x not in person.prohibited())
        
        if home is None:
            return
        
        home.move_in(person)
        person.set_home(home)
        