'''River UI parts.'''

from concepts.geometry import *
from concepts.color import *

from ui.components import *
from ui.texture import *
from ui.window import *

from building import *
from map import *
from people import *
from views import *
from panels import *
from buttons import *


class RiverWindow(PrettyWindow):
    def __init__(self, size, campaigns, image, icon):
        frame_color      = Color(255, 255, 0)
        background_color = Color(64, 64, 216)
        
        super(RiverWindow, self).__init__('Twisted River Community', icon,
                                          size, frame_color, background_color)
        
        self._image        = RiverImage(Vector(0, 0), self.surface(), image)
        self._campaigns    = campaigns
        self._scenarios    = None
        self._scenario     = None
        self._menu_visible = False
        self._menu         = MenuPanel(self.full_area().cropped(375, 282),
                                       self)
        self.add(self.image())
        self.show_menu()
    
    def image(self):
        return self._image
    
    def scenario(self):
        return self._scenario

    def scenarios(self):
        if self._scenarios is None:
            return None
        else:
            return list(self._scenarios)

    def campaigns(self):
        return list(self._campaigns)
    
    def menu(self):
        return self._menu
    
    def menu_visible(self):
        return self._menu_visible
    
    def show_menu(self):
        if self.menu_visible():
            return
        
        self._menu_visible = True
        self.menu().clear_buttons()
        self.add(self.menu())

        scenario = lambda x: lambda: self.load_scenario(x)
        campaign = lambda x: lambda: self.select_campaign(x)
        
        if self.scenario() is None:
            nr = 1

            if self.scenarios() is None:
                for xs in self.campaigns():
                    if len(xs) == 1:
                        self.menu().add_button(nr, xs[0].name(), scenario(xs[0]))
                    else:
                        self.menu().add_button(nr, 'Campaign', campaign(xs))
                        
                    nr += 1
                
                self.menu().add_button(0, 'Exit Game', self.quit)
            else:
                for x in self.scenarios():
                    self.menu().add_button(nr, x.name(), scenario(x))
                    nr += 1

                self.menu().add_button(0, 'Back', self.deselect_campaign)
        else:
            self.menu().add_button(1, 'Restart Scenario', scenario(self.scenario().__class__))
            self.menu().add_button(2, 'Abort Scenario',   self.abort_scenario)
            self.menu().add_button(3, 'Exit Game',        self.quit)
            self.menu().add_button(0, 'Continue Playing', self.hide_menu)
            self.scenario().time().stop()
        
    def hide_menu(self):
        if not self.menu_visible():
            return
        
        self._menu_visible = False
        self.remove(self.menu())
        
        if self.scenario() is not None:
            self.scenario().time().start()

    def update_menu(self):
        self.hide_menu()
        self.show_menu()
    
    def select_campaign(self, x):
        self._scenarios = x
        self.update_menu()
        
    def deselect_campaign(self):
        self._scenarios = None
        self.update_menu()
        
    def do_click(self, pos, button):
        if self.menu_visible():
            components = self.menu().components()
        else:
            components = self.all_components()
            
        for component in components:
            if not component.area().includes(Vector(*pos)):
                continue
            
            if button == 1:
                component.do_left_click(Vector(*pos) - component.origin())
            elif button == 2:
                component.do_middle_click(Vector(*pos) - component.origin())
            elif button == 3:
                component.do_right_click(Vector(*pos) - component.origin())

    def do_quitting(self):
        if self.scenario() is not None:
            self.scenario().stop()
    
    def abort_scenario(self):
        if self.scenario() is None:
            return
        
        self.scenario().stop()
        self._scenario = None
        
        self.hide_menu()
        self.remove_all()
        self.add(self.pretty())
        self.add(self.image())
        self.show_menu()
                
    def load_scenario(self, scenario_type):
        if self.scenario() is not None:
            self.abort_scenario()
            
        self.hide_menu()
        self.remove(self.image())        
            
        scenario = self._scenario = scenario_type()
        scenario.start()
        
        frame_color      = Color(255, 255, 0)
        background_color = Color(64, 64, 216).dimmed()
        border_size      = 7
        logger           = scenario.logger()
        
        client_area = self.full_area().cropped(border_size)
        left        = client_area.cropped(0, 0, 249 + border_size, 0)
        right       = client_area.with_width(249).moved(left.extent().x() + 
                                                        border_size, 0)
        view        = ViewPanel(right.cropped(0,  48 + border_size, 
                                              0, 250 + border_size),
                                scenario)
        main        = MainPanel(left.cropped(0,  48 + border_size, 
                                             0, 103 + border_size),
                                frame_color, background_color, scenario,
                                view.on_select)
        resources   = ResourcesPanel(left.with_height(48), frame_color,
                                     background_color, scenario.city())
        text        = TextPanel(left.with_height(103).moved(0, 48 + 
                                                            2 * border_size + 
                                                            main.area().extent().y()), 
                                frame_color, background_color, logger)
        button      = ButtonPanel(right.with_height(48), scenario, self)
        building    = BuildingPanel(right.with_height(250).moved(0, 48 + 
                                                                 2 * border_size + 
                                                                 view.area().extent().y()), 
                                    scenario, main)
        
        self.add(main)
        self.add(resources)
        self.add(text)
        self.add(button)
        self.add(view)
        self.add(building)

class ButtonRow(object):
    def __init__(self, panel, nr):
        self._buttons = []
        self._panel   = panel
        self._nr      = nr
        
    def buttons(self):
        return self._buttons
    
    def panel(self):
        return self._panel
    
    def nr(self):
        return self._nr
    
    def add_button(self, building):
        button_area = self.panel().client_area().cropped(6).with_width(56).with_height(56)
        button_area = button_area.moved((button_area.extent().x() + 3) * len(self.buttons()),
                                        (button_area.extent().y() + 3) * self.nr())
        
        button = BuildingButton(button_area, self.panel(), building)
        
        self.panel().add_button(button)
        self._buttons.append(button)
    
    
class BuildingPanel(RiverPanel):
    def __init__(self, area, scenario, main):
        super(BuildingPanel, self).__init__(area, scenario)
        
        self._buttons  = set()
        self._rows     = {}
        self._building = None
        self._main     = main
        
        nr = 0
        
        for category in Building.categories():
            self._rows[category] = ButtonRow(self, nr)
            
            nr += 1
            
            for type in Building.types(category):
                self._rows[category].add_button(type)
                        
    def main(self):
        return self._main
        
    def buttons(self):
        return set(self._buttons)
    
    def add_button(self, button):
        self._buttons.add(button)
        self.add(button)
        
    def set_building(self, building):
        if building == self.building():
            building = None
            
        self._building = building
        
        for button in self.buttons():
            button.highlight_if_building(building)
            
        if building is not None:
            self.city().log('%s costs: %s' % (building.name(),
                                             ', '.join(map(str, building.cost()))))
            
            self.main().set_build_callback(self.build, building)
        else:
            self.main().set_build_callback(None, None)
            
    def build(self, tile, building):
        if tile is None:
            self.set_building(None)
            return
        
        self.do_event(lambda: self.city().start_construction(tile, building))

    def building(self):
        return self._building
    
            
class BuildingButton(RiverButton):
    def __init__(self, area, panel, building):
        super(BuildingButton, self).__init__(area)
        
        self._panel    = panel
        self._building = building
    
    def panel(self):
        return self._panel
    
    def building(self):
        return self._building
    
    def highlight_if_building(self, building):
        self.set_state(building == self.building())
        
    def do_click(self, vector):
        self.panel().set_building(self.building())
        
    def render_contents(self, renderer):
        area = self.client_area().cropped(11, 4, 11, 18)
        
        renderer.draw_filled(area, self.building().color())
        renderer.draw_rectangle(area, Color(0, 0, 0))
        renderer.draw_text(self.building().name(), Vector(4, 38),
                           Color(255, 255, 255),
                           self.client_area().extent().x() - 8, True, True)
    
    def render(self):
        if not self.panel().scenario().is_allowed(self.building()):
            return
        
        super(BuildingButton, self).render()


class RiverImage(Component):
    def __init__(self, origin, surface, image):
        self._image = pygame.image.load(image)
        
        extent = Vector(self._image.get_width(),
                        self._image.get_height())
        
        super(RiverImage, self).__init__(SurfaceArea(origin, extent, surface))
    
    def image(self):
        return self._image
    
    def render(self):
        self.blit(self.image(), Vector(0, 0))
    