'''Manufactoring knowledge.'''

from concepts.color import *
from concepts.geometry import *

from river.building import *
from river.goods import *
from river.terrain import *

from river.actions.jobs import *

import random
import operator
    

class Plaza(Building):
    def __init__(self, city, tile):
        super(Plaza, self).__init__(city, tile, [Philosopher(self)])
        
        self.city().time().do_every(1, self.act)
    
    @classmethod
    def name(cls):
        return 'Plaza'
    
    @classmethod
    def stands_free(cls):
        return False
    
    @classmethod
    def size(self):
        return 3
    
    @classmethod
    def cost(self):
        return [Stone(10)]
    
    @classmethod
    def color(self):
        return Color(216, 255, 216)
    
    @classmethod
    def difficulty(self):
        return 0.5
    
    def capacity(self):
        return self.size() * self.size() - 1
    
    def full(self):
        return len(self.people()) >= self.capacity()
    
    def philosopher(self):
        return self.jobs()[0].person()
    
    def signal(self):
        if not self.full():
            x =  float(len(self.people()) - 1) / (self.capacity() - 1)
        else:
            x = float(self.capacity() + 1) / (len(self.people()) + 1)
       
        p = self.philosopher()
        
        if p is not None and p.is_at(self.office()):
            x = min(1.0, x + 0.5 * p.job().performance())
            
        return x
    
    def niveau(self):
        num = len(self.people())
        
        if num == 0:
            return 0
        
        x = reduce(operator.add,
                   map(lambda x: (x.mind() + 3 * x.spirit()) / 100.0,
                       self.people())) / num                       
       
        p = self.philosopher()
        
        if p is not None and p.is_at(self.office()):
            x = min(1.0, (x + p.job().performance() * 1.5) / 2)
        elif num < 2:
            return 0
            
        return x
                       
    
    def efficiency(self):
        return self.signal() * self.niveau()
        
    def act(self):
        for x in self.people():
            if x.social().amount() < self.efficiency():
                x.social().add(0.007 * self.efficiency())
            else:
                x.social().add(0.002 * self.efficiency())
            
                        
Building.add_type(Plaza, 'infra', 3)


class Philosopher(Job):
    def __init__(self, plaza):
        super(Philosopher, self).__init__(plaza)
        
    def title(self):
        return 'Philosopher'
    
    def color(self):
        return Color(128, 255, 128)
    
    def body_weight(self):
        return 0
    
    def mind_weight(self):
        return 1
    
    def spirit_weight(self):
        return 3
    
    def work_speed(self):
        return 0.001

    def target(self):
        return self.office()
    
    def route_to_office(self):
        return self.person().route_to(self.office())
    
    def do_work(self):
        return super(Philosopher, self).plan()
    
    def plan(self):
        b = self.building()
        p = self.person()
        
        if p is None:
            return
        
        if not p.is_at(b.office()):
            return self.route_to_office()
            
        return self.do_work()
    

class School(Building):
    def __init__(self, city, tile):
        super(School, self).__init__(city, tile, [Teacher(self),
                                                  Student(self),
                                                  Student(self),
                                                  Student(self)])
    
    @classmethod
    def name(cls):
        return 'School'
    
    @classmethod
    def size(self):
        return 2
    
    @classmethod
    def cost(self):
        return [Wood(4), Stone(4), Marble(10)]
    
    @classmethod
    def color(self):
        return Color(128, 255, 128)
    
    @classmethod
    def difficulty(self):
        return 1
    
    def teacher(self):
        p = self.jobs()[0].person()
        
        if p is None or not p.is_at(self.office()):
            return None
        else:
            return p
    
    def quality(self):
        if self.teacher() is None:
            return 0.0
        
        return self.jobs()[0].performance()
  
    def body(self):
        if self.teacher() is None:
            return 0
        
        return self.teacher().body()
  
    def mind(self):
        if self.teacher() is None:
            return 0
        
        return self.teacher().mind()
  
    def spirit(self):
        if self.teacher() is None:
            return 0
        
        return self.teacher().spirit()

    
Building.add_type(School, 'living', 2)


class Teacher(Job):
    def __init__(self, school):
        super(Teacher, self).__init__(school)
        
    def title(self):
        return 'Teacher'
    
    def color(self):
        return Color(0, 255, 0)
    
    def body_weight(self):
        return 0
    
    def mind_weight(self):
        return 3
    
    def spirit_weight(self):
        return 1
    
    def work_speed(self):
        return 0.001

    def target(self):
        return self.office()
    
    def route_to_office(self):
        return self.person().route_to(self.office())
        
    def do_work(self):
        return super(Teacher, self).plan()
    
    def plan(self):
        b = self.building()
        p = self.person()
        
        if p is None:
            return
        
        if not p.is_at(b.office()):
            return self.route_to_office()
            
        return self.do_work()


class Student(Job):
    def __init__(self, school):
        super(Student, self).__init__(school)
        
        self.restart(True)
    
    def restart(self, full):
        self._progress = 0.0
        self._current  = 0.0
        self._body     = 1
        self._mind     = 1
        self._spirit   = 1
        
        if full:
            self._old_value = 0.0
        
    def progress(self):
        return self._progress
    
    def current(self):
        return self._current
    
    def title(self):
        return 'Student'
    
    def color(self):
        return Color(0, 255, 168)

    def skill(self):
        teacher = self.building().teacher()
        
        if teacher is None or not self.person().is_inside(self.building()):
            return 0.0
        
        if self.current() > 0.0:
            value = (self.progress() * teacher.attributes().total() /
                     self.person().attributes().total() / 
                     self.person().attributes().total() / 
                     self.current()) * 30.0
        
            if value > 0.0 and self.building().teacher() is not None:
                self._old_value = value
            
        return min(self._old_value, 1.0)
    
    def work_speed(self):
        return 0.01
    
    def body_weight(self):
        return max(1, self._body)
    
    def mind_weight(self):
        return max(1, self._mind)
    
    def spirit_weight(self):
        return max(1, self._spirit)
 
    def employ(self, person):
        super(Student, self).employ(person)
        
        self.restart(True)
        
    def acted(self):
        if self.person() is None:
            return
        
        self._progress += self.work_speed() * self.building().quality()
        self._body     += self.building().body()   - self.person().body()
        self._mind     += self.building().mind()   - self.person().mind()
        self._spirit   += self.building().spirit() - self.person().spirit()
        self._current  += self.work_speed()
        
    def work_done(self):
        if random.random() < self.performance() / 10.0:
            self.person().attributes().advance(self)
            
        self.restart(False)
        