'''UI views on objects.'''

from concepts.geometry import *
from concepts.color import *

from ui.components import *
from ui.texture import *
from ui.window import *

from building import *
from buttons import *

from river.industries.knowledge import *
from river.industries.iron import *
from river.industries.mythology import *
from river.industries.storage import *
from river.industries.wine import *

import math
import random


class ObjectView(Component):
    def __init__(self, parent, area):
        super(ObjectView, self).__init__(area)

        self._parent = parent
        self._added  = []
    
    def added(self):
        return list(self._added)
    
    def remove_added(self):
        self._added = []
    
    def dispose(self):
        self.remove_added()
        super(ObjectView, self).dispose()
        
    def parent(self):
        return self._parent
    
    def add(self, element):
        self._added.append(element)
    
    def add_command_button(self, origin, extent, text, command):
        self.add(CommandButton(SurfaceArea(origin + self.parent().area().origin(),
                                           extent, self.parent().surface()),
                               self.parent(),
                               text, command))
        
    def draw_attribute(self, text, y, stat):
        size = self.font_size()

        self.draw_text(text + ':', Vector(7, y), Color(255, 255, 255), 80)
        
            
        self.draw_bar(stat.amount(), Vector(90, y),
                      Vector(self.area().extent().x() - 97, size), stat.color(),
                      Color(255, 255, 255))
        
    def draw_progress(self, text, y, amount, color):
        size = self.font_size()

        self.draw_text(text + ':', Vector(7, y), Color(255, 255, 255), 80)
        
            
        self.draw_bar(amount, Vector(90, y),
                      Vector(self.area().extent().x() - 97, size), color,
                      Color(255, 255, 255))

    def draw_label(self, text, y, value, color = Color(255, 255, 255),
                   button_text = None, command = lambda: None):
        size = self.font_size()

        self.draw_text(text + ':', Vector(7, y), Color(255, 255, 255), 80)
        self.draw_text(value, Vector(90, y), color,
                       self.area().extent().x() - 97)

        if button_text is not None:
            self.add_command_button(Vector(198, y),
                                    Vector(45, 18), button_text, command)

    def draw_line(self, text, y, color = Color(255, 255, 255),
                   button_text = None, command = lambda: None):
        size = self.font_size()

        self.draw_text(text, Vector(7, y), color,
                       self.area().extent().x() - 14)

        if button_text is not None:
            self.add_command_button(Vector(198, y),
                                    Vector(45, 18), button_text, command)
            
    def draw_resources(self, text, y, resources):
        size = self.font_size()

        self.draw_text(text + ':', Vector(7, y), Color(255, 255, 255), 80)
        
        x = 90
        
        for resource in resources:
            color = resource.color()
            nr    = str(resource.number())
            
            self.draw_text(nr, Vector(x, y), color)
            self.draw_lozengy(color, color.dimmed(), 
                              Vector(x + size / 2 * (1 + len(nr)),
                                     y + size / 2), size / 3)
    
            x += 2 * size + len(nr) / 2 * size
    
    
class PersonView(ObjectView):
    def __init__(self, parent, area, person):
        super(PersonView, self).__init__(parent, area)

        self._person = person
    
    def person(self):
        return self._person
        
    def render(self):
        pos  = 7
        size = self.font_size()
        
        self.draw_circle(self.person().color(), self.person().color().dimmed(),
                         Vector(self.area().extent().x() / 2, pos + size),
                         size, 2)
        
        pos += 7 + 2 * size
        
        self.draw_text(self.person().name(), Vector(7, pos), 
                       Color(255, 255, 255),
                       self.area().extent().x() - 14, True)
        
        pos += self.font_size() * 1.3
        
        if self.person().has_job():
            title = self.person().job().title()
        else:
            title = 'Villager'
            
        self.draw_text(title, Vector(7, pos), self.person().color(),
                       self.area().extent().x() - 14, True)
        
        pos += self.font_size() * 1.3 + 12

        self.draw_text('Body: %d    Mind: %d    Spirit: %d' %
                       (self.person().body(), self.person().mind(), 
                        self.person().spirit()),
                       Vector(7, pos), Color(255, 255, 255),
                       self.area().extent().x() - 7, True)
        
        pos += self.font_size() * 1.3 + 12
            
        pos += 8
        
        self.draw_text('Job', Vector(7, pos), 
                       Color(255, 255, 255),
                       self.area().extent().x() - 14, True)
        
        pos += size * 1.3 + 4
        
        if self.person().has_job():
            self.draw_progress('Skill', pos, self.person().skill(),
                               self.person().job().color())
        else:
            self.draw_label('Skill', pos, 'Not applicable',
                            Color(128, 128, 128))
        
        pos += self.font_size() * 1.3
        
        if self.person().has_job():
            perf  = self.person().job_performance()
            color = self.person().job().color()
            
            if perf < self.person().skill():
                color = color.dimmed()
            elif perf > self.person().skill():
                color = color.brightened()
                
            self.draw_progress('Efficiency', pos, perf, color)
        else:
            self.draw_label('Efficiency', pos, 'Not applicable',
                            Color(128, 128, 128))
        
        pos += self.font_size() * 1.3 + 12
        
        pos += 8
        
        self.draw_text('Desires', Vector(7, pos), 
                       Color(255, 255, 255),
                       self.area().extent().x() - 14, True)
        
        pos += size * 1.3 + 4            
        
        self.draw_attribute('Food', pos, self.person().nutrition())
        
        pos += self.font_size() * 1.3
        
        self.draw_attribute('Energy', pos, self.person().energy())
        
        pos += self.font_size() * 1.3
        
        self.draw_attribute('Social', pos, self.person().social())
        
        pos += self.font_size() * 1.3
        
        self.draw_attribute('Happiness', pos, self.person().happiness())
            
        pos += self.font_size() * 1.3 + 12
            
        pos += 8
        
        self.draw_text('Carrying', Vector(7, pos), 
                       Color(255, 255, 255),
                       self.area().extent().x() - 14, True)
        
        pos += size * 1.3 + 4
        
        carrying = self.person().backpack()
        
        if carrying is None:            
            self.draw_label('Resource', pos, 'Nothing', Color(128, 128, 128))
        else:
            self.draw_resources('Resource', pos, [carrying])
        
        pos += self.font_size() * 1.3 + 12
        

class BuildingView(ObjectView):
    def __init__(self, parent, area, building):
        super(BuildingView, self).__init__(parent, area)

        self._building = building
        self._method   = {ConstructionSite: self.render_construction,
                          House           : self.render_house,
                          Plaza           : self.render_plaza,
                          Winery          : self.render_processing,
                          Oven            : self.render_processing,
                          Furnace         : self.render_processing,
                          Storage         : self.render_storage,
                          Temple          : self.render_temple,
                          Mirror          : self.render_mirror}
        
    def building(self):
        return self._building
            
    def render_processing(self, b, size, pos, color):
        needed    = b.needed()
        resources = b.all_resources()
        
        pos += 8
        
        self.draw_text('Raw Material', Vector(7, pos), 
                       Color(255, 255, 255),
                       self.area().extent().x() - 14, True)
        
        pos += size * 1.3 + 4
        
        for resource in resources:
            self.draw_progress(resource.plural(), pos, b.progress_of(resource),
                               resource.color())
            
            pos += size * 1.3

        if needed:
            self.draw_resources('Needs', pos, needed)
        else:
            self.draw_label('Needs', pos, 'Nothing', Color(128, 128, 128))
            
        pos += size * 1.3
    
        return pos
    
    def render_construction(self, b, size, pos, color):
        needed = b.goods_needed()

        pos += 8
        
        self.draw_text('Construction', Vector(7, pos), 
                       Color(255, 255, 255),
                       self.area().extent().x() - 14, True)
        
        pos += size * 1.3 + 4
        
        self.draw_label('Building', pos, b.type().name(),
                        b.type().color().brightened())
        
        pos += size * 1.3
        
        self.draw_progress('Progress', pos, b.progress(), color)
        
        pos += size * 1.3

        if needed:
            self.draw_resources('Needs', pos, needed)
        else:
            self.draw_label('Needs', pos, 'Nothing', Color(128, 128, 128))
            
        pos += size * 1.3
    
        return pos
    
    def render_house(self, b, size, pos, color):
        pos += 8
        
        self.draw_text('Inhabitants', Vector(7, pos), 
                       Color(255, 255, 255),
                       self.area().extent().x() - 14, True)
        
        pos += size * 1.3 + 4
            
        allow  = lambda r: lambda: r.allow_filling()
        forbid = lambda r: lambda: r.forbid_filling()
        evict  = lambda p: lambda: b.evict(p)
        
        for r in b.rooms():
            p = r.occupant()
            s = r.must_not_fill()
            
            if s:
                self.draw_line('Must not fill', pos,
                                Color(255, 128, 128),
                                'Allow', allow(r))
            elif p is None:
                self.draw_line('Not filled', pos,
                                Color(128, 128, 128),
                                'Forbid', forbid(r))
            else:
                text = '%s (%s)' % (p.name(), p.job_description())
                self.draw_line(text, pos, p.color().brightened(),
                               'Evict', evict(p))
        
            pos += size * 1.3
    
        return pos
            
    def render_plaza(self, b, size, pos, color):
        pos += 8
        
        self.draw_text('Conversation', Vector(7, pos), 
                       Color(255, 255, 255),
                       self.area().extent().x() - 14, True)
        
        pos += size * 1.3 + 4
        
        self.draw_progress('Signal', pos, b.signal(), color)
            
        pos += size * 1.3
        
        self.draw_progress('Niveau', pos, b.niveau(), color)
            
        pos += size * 1.3
        
        self.draw_progress('Quality', pos, b.efficiency(), color)
        
        pos += size * 1.3
    
        return pos
    
    def render_storage(self, b, size, pos, color):
        resources = ResourceOverview(b.city().goods())

        pos += 8
        
        self.draw_text('Policy', Vector(7, pos), 
                       Color(255, 255, 255),
                       self.area().extent().x() - 14, True)
        
        pos += size * 1.3 + 4
        
        allow  = lambda x: lambda: b.allow(x)
        forbid = lambda x: lambda: b.forbid(x)
        
        for r in resources.types():
            s = b.policy_of(r)
            
            if s:
                self.draw_label(r.plural(), pos, 'Allowed',
                                r.color().brightened(),
                                'Forbid', forbid(r))
            else:
                self.draw_label(r.plural(), pos, 'Forbidden',
                                Color(255, 128, 128),
                                'Allow', allow(r))
        
            pos += size * 1.3
            
        return pos
            
    def render_temple(self, b, size, pos, color):
        pos += 8
        
        self.draw_text('God', Vector(7, pos), 
                       Color(255, 255, 255),
                       self.area().extent().x() - 14, True)
        
        pos += size * 1.3 + 4
        
        self.draw_progress('Anger', pos, b.anger(), color.brightened())
            
        pos += size * 1.3

        self.draw_progress('Danger', pos, b.danger(), color.brightened())
            
        pos += size * 1.3
    
        return pos

            
    def render_mirror(self, b, size, pos, color):
        needed    = b.needed()
        resources = b.all_resources()
        
        pos += 8
        
        self.draw_text('Aqua', Vector(7, pos), 
                       Color(255, 255, 255),
                       self.area().extent().x() - 14, True)
        
        pos += size * 1.3 + 4
        
        for resource in resources:
            self.draw_progress('Power', pos, b.progress_of(resource),
                               resource.color())
            
            pos += size * 1.3

        if needed:
            self.draw_resources('Remaining', pos, needed)
        else:
            self.draw_label('Remaining', pos, 'Nothing', Color(128, 128, 128))
            
        pos += size * 1.3
    
        return pos

    def render(self):
        self.remove_added()
        
        pos   = 7
        size  = self.font_size()
        b     = self.building()
        color = b.color()
        
        self.draw_filled(SurfaceArea(Vector(self.area().extent().x() / 2 - size,
                                            pos) + self.origin(),
                                     Vector(size * 2, size * 2),
                                     self.surface()), color)
        self.draw_rectangle(SurfaceArea(Vector(self.area().extent().x() / 2 - size,
                                               pos) + self.origin(),
                                        Vector(size * 2, size * 2),
                                        self.surface()), color.dimmed())
        
        pos += 7 + 2 * size
        
        self.draw_text(b.name(), Vector(7, pos), 
                       Color(255, 255, 255),
                       self.area().extent().x() - 14, True)
        
        pos += size * 1.3 + 12
        
        if b.__class__ in self._method:
            pos  = self._method[b.__class__](b, size, pos, color)        
            pos += 12
            
        pos += 8
        
        self.draw_text('Storage', Vector(7, pos), 
                       Color(255, 255, 255),
                       self.area().extent().x() - 14, True)
        
        pos += size * 1.3 + 4
        
        resources = ResourceOverview(list(b.resources()))

        if resources.is_empty():
            self.draw_label('Stored', pos, 'Nothing', Color(128, 128, 128))
        else:
            self.draw_resources('Stored', pos, resources.list())
        
        pos += size * 1.3 + 12
        
        if b.jobs():
            pos += 8
            
            self.draw_text('Jobs', Vector(7, pos), 
                           Color(255, 255, 255),
                           self.area().extent().x() - 14, True)
            
            pos += size * 1.3 + 4
            
            allow  = lambda job: lambda: job.allow_filling()
            forbid = lambda job: lambda: job.forbid_filling()
            fire   = lambda job: lambda: job.fire_worker()
            
            for job in b.jobs():
                p = job.person()
                s = job.must_not_fill()
                
                if s:
                    self.draw_label(job.title(), pos, 'Must not fill',
                                    Color(255, 128, 128),
                                    'Allow', allow(job))
                elif p is None:
                    self.draw_label(job.title(), pos, 'Not filled',
                                    Color(128, 128, 128),
                                    'Forbid', forbid(job))
                else:
                    text = '%s (%d%%)' % (p.name(), p.job_performance() * 100)
                    self.draw_label(job.title(), pos, text,
                                    job.color().brightened(),
                                    'Fire', fire(job))
            
                pos += size * 1.3
                    
        
class TileView(ObjectView):
    def __init__(self, parent, area, tile):
        super(TileView, self).__init__(parent, area)

        self._tile = tile
        
    def tile(self):
        return self._tile            
    
    def render(self):
        pos   = 7
        size  = self.font_size()
        t     = self.tile()
        color = t.terrain().color()
        
        self.draw_filled(SurfaceArea(Vector(self.area().extent().x() / 2 - size,
                                            pos) + self.origin(),
                                     Vector(size * 2, size * 2),
                                     self.surface()), color)
        self.draw_rectangle(SurfaceArea(Vector(self.area().extent().x() / 2 - size,
                                               pos) + self.origin(),
                                        Vector(size * 2, size * 2),
                                        self.surface()), color.dimmed())
        
        pos += 7 + 2 * size
        
        self.draw_text(str(t.terrain()), Vector(7, pos), 
                       Color(255, 255, 255),
                       self.area().extent().x() - 14, True)
        
        pos += size * 1.3 + 12
            
        pos += 8
        
        self.draw_text('Ground', Vector(7, pos), 
                       Color(255, 255, 255),
                       self.area().extent().x() - 14, True)
        
        pos += size * 1.3 + 4
        
        resource = t.goods()

        if resource is None:
            self.draw_label('Resource', pos, 'None', Color(128, 128, 128))
        else:
            self.draw_resources('Resource', pos, [resource])

    
    
class CityView(ObjectView):
    def __init__(self, parent, area, city):
        super(CityView, self).__init__(parent, area)

        self._city = city
    
    def city(self):
        return self._city
        
    def render(self):
        pos     = 7
        size    = self.font_size()
        persons = list(self.city().persons())
        persons.sort(lambda a, b: cmp(a.name(), b.name()))
        
        self.draw_text(self.city().name(), Vector(7, pos), 
                       Color(128, 192, 255),
                       self.area().extent().x() - 14, True)
        
        pos += self.font_size() * 1.3
        
        self.draw_text('%d Inhabitants' % len(persons), Vector(7, pos), 
                       Color(255, 255, 255),
                       self.area().extent().x() - 14, True)
        
        pos += self.font_size() * 1.3 + 8
        
        color = Color(222, 222, 128)
        
        self.draw_text('Name',  Vector(  7, pos), color, 70)
        self.draw_text('Skill', Vector( 80, pos), color, 40, True)
        self.draw_text('B',     Vector(120, pos), color, 25, True)
        self.draw_text('M',     Vector(145, pos), color, 25, True)
        self.draw_text('S',     Vector(170, pos), color, 25, True)
        self.draw_text('Happy', Vector(195, pos), color, 45, True)

        pos += self.font_size() * 1.3 + 4
        
        def attribute_color(x):
            if x <  5:
                return Color(255, 128, 128)
            if x < 10:
                return Color(255, 255, 128)
            if x < 15:
                return Color(255, 255, 255)
            if x < 20:
                return Color(128, 255, 128)
            if x <  25:
                return Color(0, 192, 0)

            return Color(0, 192, 255)
        
        def happy_color(x):
            if x < 0.10:
                return Color(255, 0, 0)
            if x < 0.25:
                return Color(192, 0, 0)
            if x > 0.90:
                return Color(0, 255, 0)
            if x > 0.75:
                return Color(0, 192, 0)

            return Color(255, 255, 255)
            
        for person in persons:
            color  = person.color().brightened()
            name   = person.name()
            skill  = '%d%%' % (person.skill() * 100)
            b      = str(person.body())
            m      = str(person.mind())
            s      = str(person.spirit())
            happy  = '%d%%' % (person.happiness().amount() * 100)
            hcolor = happy_color(person.happiness().amount())
            acolor = lambda x: attribute_color(int(x))
            
            self.draw_text(name,  Vector(  7, pos), color,     70)
            self.draw_text(skill, Vector( 77, pos), color,     40, None)
            self.draw_text(b,     Vector(117, pos), acolor(b), 25, None)
            self.draw_text(m,     Vector(142, pos), acolor(m), 25, None)
            self.draw_text(s,     Vector(167, pos), acolor(s), 25, None)
            self.draw_text(happy, Vector(192, pos), hcolor,    45, None)
            
            pos += self.font_size() * 1.3
        