'''GUI Panels.'''

from concepts.geometry import *
from concepts.color import *

from ui.components import *
from ui.texture import *
from ui.window import *

from building import *
from map import *
from people import *
from views import *
from objectviews import *
from gui import *
from buttons import *


class RiverLogo(Component):
    def __init__(self, origin, surface):
        self._image = pygame.image.load(self.logo())
        
        extent = Vector(self._image.get_width(),
                        self._image.get_height())
        
        super(RiverLogo, self).__init__(SurfaceArea(origin, extent, surface))
    
    def image(self):
        return self._image
    
    def render(self):
        self.blit(self.image(), Vector(0, 0))
        self.draw_rectangle(self.area(), Color(255, 255, 0))
        

class RiverPanel(Panel):
    def __init__(self, area, scenario):
        super(RiverPanel, self).__init__(area, Color(255, 255, 0),
                                          Color(64, 64, 216).dimmed())
        
        self._scenario = scenario
    
    def scenario(self):
        return self._scenario
    
    def city(self):
        return self.scenario().city()
    
    def tile(self, x, y):
        return self.scenario().map().tile(x, y)
    
    def do_event(self, event):
        self.scenario().time().add_ui_event(event)


class MenuPanel(RiverPanel):
    def __init__(self, area, window):
        super(MenuPanel, self).__init__(area, window.scenario())
    
        self.add(RiverLogo(area.origin() + Vector(4, 4), self.surface()))
        
        self._buttons = []

    def buttons(self):
        return list(self._buttons)
    
    def clear_buttons(self):
        for x in self.buttons():
            self.remove(x)
            
        self._buttons = []
        
    def add_button(self, pos, text, command):
        if pos == 0:
            pos  = 5.25
        else:
            pos += 1
        
        x = CommandButton(self.client_area().with_height(32).moved(0, pos * 32),
                          self, text, command, False, True)
        
        self._buttons.append(x)    
        self.add(x)


class MainPanel(Panel):
    def __init__(self, area, frame_color, background_color, scenario, 
                 on_select):
        super(MainPanel, self).__init__(area, frame_color, background_color)
        
        self._view = MapView(self.client_area(), scenario.city(),
                             15, Color(255, 255, 255), on_select)
        
        self.add(self.view())

    def view(self):
        return self._view
    
    def set_build_callback(self, callback, building):
        self.view().set_build_callback(callback, building)


class ResourcesPanel(Panel):
    def __init__(self, area, frame_color, background_color, city):
        super(ResourcesPanel, self).__init__(area, frame_color,
                                             background_color)
    
        self.add(ResourcesView(self.client_area(), city))


class ButtonPanel(RiverPanel):
    def __init__(self, area, scenario, window):
        super(ButtonPanel, self).__init__(area, scenario)

        def name():
            if scenario.is_won():
                return 'Continue'
            else:
                return 'Objectives'
        
        def action():
            if not scenario.is_won():
                scenario.print_objectives()
                return
            
            next = scenario.next_scenario()
            
            if next is None:
                window.abort_scenario()
            else:
                window.load_scenario(next)
                
        half = self.client_area().extent().x() / 2
        
        self.add(CommandButton(self.client_area().with_width(half),
                               self, 'Menu', window.show_menu, False, True))

        self.add(CommandButton(self.client_area().with_width(half).moved(half + 1, 0), 
                               self, name, action, False))


class ViewPanel(RiverPanel):
    def __init__(self, area, scenario):
        super(ViewPanel, self).__init__(area, scenario)
        
        self._view = None
        
        self.show_view(CityView(self, self.client_area(), scenario.city()))
        
    def view(self):
        return self._view
    
    def show_view(self, view):
        if self._view is not None:
            self.remove(self._view)
            self._view.dispose()
            
        self._view = view
        
        if view is not None:
            self.add(view)
        
    def on_select(self, thing):
        if thing is None:
            self.show_view(CityView(self, self.client_area(),
                                    self.scenario().city()))
        elif thing.__class__ is Person:
            self.show_view(PersonView(self, self.client_area(), thing))
        elif thing.__class__ is Tile:
            self.show_view(TileView(self, self.client_area(), thing))            
        else:
            self.show_view(BuildingView(self, self.client_area(), thing))
    
    def real_components(self):
        return super(ViewPanel, self).components()
    
    def components(self):
        if not self.view() is not None:
            return self.real_components()
        
        return self.real_components() + self.view().added()
    
    def render(self):
        for component in self.real_components():
            component.render()
            
        if self.view() is not None:
            for component in self.view().added():
                component.render()
                
        
class TextPanel(Panel):
    def __init__(self, area, frame_color, background_color, logger):
        super(TextPanel, self).__init__(area, frame_color,
                                        background_color)
        
        self.add(TextView(self.client_area(), logger))
