'''Huggable people in river. :-)'''

from concepts.color import *
from concepts.geometry import *

from actions.consume import *
from actions.move import *
from actions.sleep import *

from industries.knowledge import *

from attributes import *
from building import *
from names import *

import random
import math


subject = object


class Person(subject):
    def __init__(self, name, city, tile):
        if name is None:
            while city.has_name(name):
                name = random.choice(NAMES)
            
        self._name       = name
        self._city       = city
        self._tile       = tile
        self._action     = None
        self._plan       = None
        self._home       = None
        self._job        = None
        self._nutrition  = Nutrition(self, 0.8)
        self._energy     = Energy(self, 1.0)
        self._social     = Social(self, 1.0)
        self._happiness  = Happiness(self, 0.5)
        self._attributes = Attributes()
        self._start_tile = tile
        self._backpack   = None
        self._priority   = 0
        self._prohibited = Prohibition(self.time())
        
        self.time().do_every(1, self.act)
    
    def prohibited(self):
        return self._prohibited
        
    def name(self):
        return self._name
    
    def color(self):
        if self.has_job():
            return self.job().color()
        
        return Color(255, 255, 255)
    
    def priority(self):
        return self._priority
    
    def base_speed(self):
        return (  0.060
                + 0.007 * self.body()
                - 0.002 * abs(self.nutrition().amount() - 0.5) * 10
                + 0.004 * min(0, self.energy().amount() - 0.5) * 10)
    
    def backpack(self):
        return self._backpack
    
    def speed(self):
        return self.base_speed() / self.tile().difficulty()
    
    def city(self):
        return self._city
    
    def time(self):
        return self.city().time()
    
    def attributes(self):
        return self._attributes
    
    def nutrition(self):
        return self._nutrition
    
    def energy(self):
        return self._energy
    
    def social(self):
        return self._social
    
    def happiness(self):
        return self._happiness
    
    def body(self):
        return self.attributes().body()
    
    def mind(self):
        return self.attributes().mind()
    
    def spirit(self):
        return self.attributes().spirit()
    
    def job(self):
        return self._job
    
    def has_job(self):
        return (self.job() is not None)
    
    def seek_employment(self):
        self.city().jobs().seek_employment(self)
    
    def seek_home(self):
        self.city().seek_home(self)
        
    def employed(self, job):
        self._job = job
        
    def fired(self, job):
        self._job = None
        self.prohibited().prohibit(job.building())
        self.happiness().reduce(0.2)
        
        if self.priority() == 2:
            self.set_plan(None, 0)
        
    def evicted(self, home):
        self._home = None
        self.prohibited().prohibit(home)
        self.happiness().reduce(0.5)
        
        if self.priority() == 1:
            self.set_plan(None, 0)
            
    def skill(self):
        job = self.job()
        
        if job is None:
            return 0.0
        
        return job.skill()
            
    def job_performance(self):
        job = self.job()
        
        if job is None:
            return 0.0
        
        return job.performance()
    
    def job_description(self):
        if self.job() is None:
            return 'Villager'
        
        return self.job().title()
        
    def home(self):
        return self._home
    
    def set_home(self, home):
        self._home = home
    
    def has_home(self):
        return (self._home is not None)
    
    def find_resource(self, resource):
        predicate = lambda g, t: g.is_at_least(resource)
        
        return self.city().nearest_tile_with_good(self, predicate)
    
    def find_matching_resource(self, p):
        predicate = lambda g, t: p(g)
        
        return self.city().nearest_tile_with_good(self, predicate)
    
    def find_matching_resource_tile(self, predicate):
        return self.city().nearest_tile_with_good(self, predicate)
    
    def find_nearest(self, p):
        return self.city().nearest_tile_with(self, p)
    
    def consume(self, number, predicate, priority = 1):
        self.set_plan(GoAndConsume(self, number, predicate), priority)
        
    def tile(self):
        return self._tile
    
    def set_tile(self, tile):
        self._tile = tile
        
    def start_tile(self):
        return self._start_tile
    
    def move_by(self, vector):
        vector += self.tile().vector()
        
        self.set_tile(self.city().map().tile(vector.x(), vector.y()))
    
    def route_to(self, tile):
        return Route(self, tile)
    
    def take(self, resource = None):
        if resource is None:
            resource = self.tile().goods()
            
        if self._backpack is not None:
            return
        
        if self.tile().goods() is None:
            return
        
        if not self.tile().goods().matches(resource):
            return
        
        self._backpack = self.tile().goods().take(resource.number())        
        self._backpack.on_destruction('person', lambda x: self.drop_or_discard())
        self.city().add_good(self._backpack)
    
    def can_drop(self):
        return self.can_drop_at(self.tile())
    
    def can_drop_at(self, tile):
        if self._backpack is None:
            return False
                
        goods = tile.goods()
        
        if goods is None:
            return True
        
        return (goods.can_take(self._backpack) > 0)
        
    def drop(self):
        if self._backpack is None:
            return
                
        goods = self.tile().goods()
        
        if goods is None:
            self.drop_or_discard()
            return
        
        if goods.matches(self._backpack):
            self.drop_as_much_as_fits()
            return
        
        (goods, self._backpack) = (self._backpack, None)
        
        goods.remove_callback('person')
        self.take()
        self.tile().set_goods(goods)
        
    def drop_as_much_as_fits(self):
        if self._backpack is None:
            return
                
        goods = self.tile().goods()
        
        if goods is None:
            self.drop_or_discard()
            return
        
        number = goods.can_take(self._backpack)
        
        goods.increase(number)
        self._backpack.decrease(number)
        
    def drop_or_discard(self):
        if self._backpack is None:
            return
        
        self._backpack.remove_callback('person')
        self.city().remove_good(self._backpack)
        
        if self._backpack.number() <= 0:
            self._backpack = None
            return
        
        goods = self.tile().goods()
        
        if goods is None:
            self.tile().set_goods(self._backpack)
        elif goods.__class__ is self._backpack.__class__:
            goods.increase(self._backpack.number())
            self._backpack.destroy()
        else:
            self._backpack.destroy()
            
        self._backpack = None
        
    def is_at(self, tile):
        return (self.tile() == tile)
        
    def is_inside(self, building):
        if building is None:
            return False
            
        return (self.tile() in building.tiles())
    
    def action(self):
        return self._action
    
    def has_action(self):
        return (self.action() is not None)
    
    def home_tile(self):
        return random.choice(list(self.home().tiles()))

    def decide_plan(self):
        if self.backpack() is not None:
            self.get_your_hands_free()
            return
        
        if self.energy().is_sleepy():
            self.sleep()
            return
        
        if self.nutrition().is_starving() and random.randint(0, 25) > self.mind():
            self.panic()
            return

        if self.nutrition().is_hungry():
            self.go_eating()
            return
        
        if not self.has_home():
            self.seek_home()
        
        if self.energy().is_tired():
            if self.has_home():
                self.sleep(self.home_tile())
                return
            else:
                self.sleep(self.find_nearest(lambda x: x.building() is None and
                                             x.difficulty() <= 1))
                return
            
        if self.social().is_lonely() or (random.random() < 0.1 and
                                         random.random() > self.social().amount()):
            if self.go_to_plaza():
                return
            
        if not self.has_job():
            self.seek_employment()
        
        if (self.has_job() and self.job().has_plan() and
            (random.random() < 0.90 or 
             random.random() < self.happiness().amount())):
            self.go_working()
            return
        
        if self.has_home():
            self.go_home()
            return
        
        self.go_wandering()
    
    def get_your_hands_free(self):
        self.set_plan(GetHandsFree(self), 200)

    def sleep(self, tile = None):
        if tile is None:
            tile = self.tile()
            
        if self.energy().is_sleepy():
            self.set_plan(GoAndSleep(self, tile), 7)
        else:
            self.set_plan(GoAndSleep(self, tile), 5)
    
    def panic(self):    
        self.set_plan(Panic(self), 100)
        
    def go_eating(self):
        if self.nutrition().is_starving():
            self.consume(10, lambda x: x.is_edible(), 8)
        else:
            self.consume(10, lambda x: x.is_edible(), 6)
    
    def go_working(self):
        self.set_plan(self.job().plan(), 2)
    
    def go_wandering(self):
        if random.randint(1, 10) == 1:
            buildings = self.city().buildings()
            
            if not buildings:
                return
            
            building = random.choice(list(buildings))
            tile     = random.choice(list(building.tiles()))
        else:
            tile = random.choice(filter(lambda x: x.exists(),
                                        self.tile().weighted_neighbors()))
            
        self.set_plan(Route(self, tile), 1)

    def go_home(self):
        self.set_plan(WanderBuilding(self, self.home(),
                                     lambda: random.random() < 0.25), 1)

    def go_to_plaza(self):
        building = self.nearest(Plaza)
        
        if not building:
            return False
        
        def condition():
            return ((random.random() < 0.03 and 
                     random.random() > building.efficiency() / 2 and 
                     random.random() < self.social().amount() and 
                     not self.social().is_lonely()) or
                     self.nutrition().is_hungry() or
                     self.energy().is_tired())
            
        self.set_plan(WanderBuilding(self, building, condition), 3)
        
        return True
        
    def nearest(self, type):
        return self.city().nearest_building_of_type(self, type)
        
    def act(self):
        self.nutrition().reduce(0.00007)
        self.energy().reduce(0.0001)
        self.social().reduce(0.0001)
        self.social().spend_time_with(self.city().persons_on(self.tile()))
        self.happiness().normalize(0.00005)
        self.happiness().calculate(0.0004)
        
        if not self.has_action():
            if not self.has_plan() or random.random() < 0.01:
                self.decide_plan()

            if not self.has_plan():
                return
            
            if not self.has_action():
                self.done_acting()
                return
            
            self._action = self.plan().next_action()
        
        action = self.action()
        
        if action is None:
            self.set_plan(None, 0)
            return
        
        if self.has_job() and self.priority() == 2:
            body = float(self.job().body_weight()) / (self.job().body_weight() +
                                                      self.job().mind_weight() + 
                                                      self.job().spirit_weight())
            
            self.nutrition().reduce(0.0002 * body)
            self.energy().reduce(0.00005 * body)
            
            if (self.is_inside(self.job().building()) and
                random.random() < self.attributes().advancement_chance()):
                self.attributes().advance(self.job())
        else:
            self.nutrition().reduce(0.00009)
            self.energy().reduce(0.00003)
            
        action.act()
        
    def done_acting(self):
        if self.has_plan():
            self._action = self.plan().next_action()
            
            if not self.has_action():
                self.set_plan(None, 0)
        else:
            self._action = None
            
        self._start_tile = self.tile()
        
    def plan(self):
        return self._plan
    
    def has_plan(self):
        return (self.plan() is not None)
    
    def set_plan(self, plan, priority = 1):
        if self.priority() >= priority and priority > 0:
            return
        
        self._plan     = plan
        self._priority = priority
        
    def displacement(self):
        action = self.action()
        
        if action is None:
            return Vector(0, 0)
        
        return self.action().displacement()

    def __str__(self):
        return ('%s (carrying: %s; job: %s; plan: %s; action: %s)' % 
                (self.name(), self.backpack(), self.job(),
                 self.plan(), self.action()))


def _above(n):
    return lambda self: self.amount() > n

def _below(n):
    return lambda self: self.amount() < n

def _between(n, m):
    return lambda self: n < self.amount() < m


class Stat(object):
    def __init__(self, person, amount):
        self._person = person
        self._amount = amount

    def person(self):
        return self._person
    
    def amount(self):
        return self._amount
    
    def set_amount(self, n):
        self._amount = n
        
    def add(self, amount):
        self._amount = max(min(self.amount() + amount, 1), 0)
    
    def reduce(self, amount):
        self._amount = max(min(self.amount() - amount, 1), 0)

    def color(self):
        if self.amount() > 0.75:
            return Color(0, 128, 0)
        
        if self.amount() > 0.50:
            return Color(0, 216, 0)
        
        if self.amount() < 0.10:
            return Color(128, 0, 0)
            
        if self.amount() < 0.25:
            return Color(255, 128, 0)
        
        return Color(255, 255, 0)

    
class Nutrition(Stat):
    is_full     = _above(0.75)
    is_satiated = _above(0.50)
    is_hungry   = _below(0.25)
    is_starving = _below(0.10)
    
    def consumption_needed(self, good):
        return int((1 - self.amount()) / good.nutrional_value())
        
        
class Energy(Stat):
    is_sleepy = _below(0.05)
    is_tired  = _below(0.25)


class Social(Stat):
    is_lonely = _below(0.25)
    
    def spend_time_with(self, people):
        value = 0
        
        for x in people:
            if self.person() == x:
                continue
            
            if random.randint(0, x.spirit()) < 2:
                value -= 0.0003
            
            delta  = max(0, x.spirit() - max(0, self.amount() - 0.25) * 20)
            value += delta * 0.00015
                
        value /= math.sqrt(len(people))
        
        self.add(value)
        
        
class Happiness(Stat):
    is_unhappy = _below(0.25)
    is_happy   = _above(0.75)
    
    def normalize(self, n):
        if self.amount() > 0.4999 and self.amount() < 0.5001:
            self.set_amount(0.5)
        elif self.amount() > 0.5:
            self.reduce(n)
        else:
            self.add(n)
    
    def delta(self, stat):
        delta = 0
        
        if stat.amount() < 0.25:
            delta += -0.25 + stat.amount()
            
        if stat.amount() < 0.10:
            delta += -0.25 + stat.amount()
            
        if stat.amount() > 0.75:
            delta += (-0.75 + stat.amount()) / 2
            
        if stat.amount() > 0.90:
            delta += (-0.75 + stat.amount()) / 2
        
        return delta
    
    def calculate(self, n):
        delta  = 0
        delta += self.delta(self.person().nutrition())
        delta += self.delta(self.person().energy())
        delta += self.delta(self.person().social())
        
        self.add(delta * n)
    
    def modifier(self):
        if self.amount() < 0.25:
            return self.amount() * 4
        
        if self.amount() > 0.75:
            return 1 + (self.amount() - 0.75) * 2
        
        return 1.0
        
        
class Prohibition(object):
    def __init__(self, time):
        self._time       = time
        self._prohibited = set()
    
    def time(self):
        return self._time
        
    def is_prohibited(self, x):
        return (x in self._prohibited)
    
    def prohibit(self, x):
        if self.is_prohibited(x):
            return
        
        self._prohibited.add(x)
        self.time().do_after(250, lambda: self.allow(x))
    
    def allow(self, x):
        self._prohibited.discard(x)
    
    def __iter__(self):
        return iter(self._prohibited)
    
    def __str__(self):
        return str(self._prohibited)
    