'''Scenario base classes.'''

from concepts.logger import *
from concepts.geometry import *

from city import *
from map import *
from time import *
from goods import *

from river.building import *
from river.industries.farming import *
from river.industries.logging import *
from river.industries.quarrying import *
from river.industries.mining import *
from river.industries.storage import *
from river.industries.knowledge import *
from river.industries.mythology import *
from river.industries.wine import *
from river.industries.iron import *

import random


class Scenario(object):
    def __init__(self):
        self._city       = None
        self._map        = None
        self._time       = Time(0.02)
        self._allowed    = set()
        self._is_won     = False
        self._objectives = []
        self._logger     = Logger()
        
    @classmethod
    def name(cls):
        return 'Unnamed Scenario'
    
    @classmethod
    def next_scenario(cls):
        return None
    
    def is_won(self):
        return self._is_won
    
    def win(self):
        self._is_won = True
        
    def objective(self, strings, color = Color(255, 255, 0)):
        self.city().log(''.join(strings), color)
        self.reset_objectives()
        self.add_objective(strings, True)
        
    def message(self, strings, color = Color(128, 128, 0)):
        self.city().log(''.join(strings), color)
        
    def hint(self, strings, color = Color(128, 128, 128)):
        self.city().log('  ' + ''.join(strings), color)
        self.add_objective(strings, False)
    
    def notice(self, strings, color = Color(0, 192, 255)):
        self.city().log(''.join(strings), color)
    
    def flavor(self, strings, color = Color(255, 255, 255)):
        self.city().log('  ' + ''.join(strings), color)
        
    def is_allowed(self, type):
        return not self._allowed or type in self._allowed

    def allow(self, type):
        self._allowed.add(type)
        
    def time(self):
        return self._time
    
    def map_size(self):
        return Vector(50, 39)
        
    def new_map(self):
        return CityMap(self.map_size())
    
    def map(self):
        if self._map is None:
            self._map = self.new_map()
    
        return self._map
    
    def city_name(self):
        return None
    
    def new_city(self):
        return City(self.city_name(), self.map(), self.time())
    
    def city(self):
        if self._city is None:
            self._city = self.new_city()
    
        return self._city
    
    def layout(self):
        return []
    
    def terrain(self):
        return TerrainMap(self.layout())
    
    def create_immigrants(self, num):
        for _ in range(num):
            self.city().give_birth(self.starting_tile())
            
    def setup(self):
        self.create_immigrants(4)
        
    def persons(self, type = None):
        if type is None:
            return self.city().persons()
        else:
            return self.city().persons_of_type(type)
    
    def buildings(self, type = None):
        if type is None:
            return self.city().buildings()
        else:
            return self.city().buildings_of_type(type)
    
    def immigrants(self):
        self.create_immigrants(2)
        self.notice('2 immigrants arrived!')
        
    def can_attract_immigrants(self):
        num = len(self.persons())
        
        return (num < 20 and len(self.buildings(House)) * 2 > num)
    
    def events(self):
        return [(self.can_attract_immigrants, self.immigrants)]

    def schedule(self):
        return {}
    
    def intro(self):
        pass
    
    def script(self):
        return []
    
    def starting_tile(self):
        return self.map().tile(0, 0)
    
    def logger(self):
        return self._logger
    
    def start(self):
        self.city().set_logger(self.logger())
        self.terrain().write_to(self.map())
        self.setup()
        self.notice(self.city_name())
        self.intro()
            
        for time in self.schedule():
            for event in self.schedule()[time]:
                self.time().do_at(time, event)
            
        for condition, event in self.script():
            self.time().do_when(condition, event)
            
        for condition, event in self.events():
            self.time().do_when(condition, event, False)
                
        self.time().start()
    
    def stop(self):
        self.time().stop()
        
    def objectives(self):
        return self._objectives
    
    def reset_objectives(self):
        self._objectives = []
        
    def add_objective(self, text, is_message):
        self._objectives.append((text, is_message))
    
    def print_objectives(self):
        objectives = self.objectives()
        
        self.reset_objectives()
        
        for text, is_message in objectives:
            if is_message:
                self.objective(text)
            else:
                self.hint(text)
            

class TerrainMap(object):
    SYMBOLS = {'.': (Grass,),
               ':': (Grass, Stone(2)),
               '"': (Grass, Wheat(2)),
               '&': (Grass, Sapphire(1)),
               ';': (Grass, Grapes(2)),
               '?': (Grass, Wine(5)),
               '~': (River,),
               '#': (Waterfall,),
               '$': (Forest,),
               '=': (Forest, Wood(5)),
               '%': (Rock,),
               '*': (Rock, Stone(5)),
               '+': (MarbleRock,),
               '_': (MarbleRock, Marble(1)),
               '^': (Mountain,),
               '!': (Mountain, None, None, lambda x: x.terrain().deposit(Coal)),
               'r': (Grass, None, Road),
               'p': (Grass, None, Plaza),
               'h': (Grass, None, House),
               'c': (Grass, None, School),
               's': (Grass, None, Storage),
               'u': (Grass, None, Furnace),
               'T': (Grass, None, Temple),
               'b': (River, None, Bridge),
               '/': (Waterfall, None, Mirror),
               'm': (Mountain, None, Mine),
               'M': (Mountain, None, Mine, lambda x: x.terrain().deposit(Coal))}
    
    def __init__(self, strings):
        self._strings = strings
    
    def character_at(self, pos):
        pos = pos.vector()
        
        return self._strings[pos.y()][pos.x()]
    
    def tuple_at(self, pos):
        return self.SYMBOLS[self.character_at(pos)]
    
    def value_at(self, pos, nr):
        all = self.tuple_at(pos)
        
        if len(all) <= nr:
            return None
        
        return all[nr]
    
    def terrain_at(self, pos):
        return self.value_at(pos, 0)
        
    def resource_at(self, pos):
        return self.value_at(pos, 1)
        
    def building_at(self, pos):
        return self.value_at(pos, 2)
        
    def code_at(self, pos):
        return self.value_at(pos, 3)

    def vectors(self):
        for y in range(len(self._strings)):
            for x in range(len(self._strings[y])):
                yield Vector(x, y)
    
    def tiles(self, map):
        for v in self.vectors():
            yield map.tile(v.x(), v.y())
        
    def write_to(self, city_map):
        for tile in self.tiles(city_map):
            terrain  = self.terrain_at(tile)
            resource = self.resource_at(tile)
            code     = self.code_at(tile)
            
            tile.become(terrain)
            
            if resource is not None:
                tile.spawn(resource)
                
            if code is not None:
                code(tile)
                
        for tile in self.tiles(city_map):
            building = self.building_at(tile)
            
            if building is not None:
                tile.build(building)
            