'''Basic UI views.'''

from concepts.geometry import *
from concepts.color import *

from ui.components import *
from ui.texture import *
from ui.window import *

from building import *

import math
import random
    
    
class TextView(Component):
    def __init__(self, area, logger):
        super(TextView, self).__init__(area)

        self._logger = logger
        
    def logger(self):
        return self._logger
    
    def render(self):
        pos  = 6
        last = self.logger().last(5)
        
        for line, color in last:
            self.draw_text(line, Vector(5, pos), color.brightened(),
                           self.area().extent().x() - 10)
            pos += self.font_size() * 1.3


class ResourcesView(Component):
    def __init__(self, area, city):
        super(ResourcesView, self).__init__(area)

        self._city = city
        
    def city(self):
        return self._city
    
    def render(self):
        resources = ResourceOverview(self.city().goods())
        size      = self.font_size()
        
        nr = 0
        
        for type in resources.types():
            vector  = Vector(5 + nr * 76, 4)
            color   = type.color()
            num     = str(resources.number_of(type))
            
            if nr == 0:
                vector += Vector(5, 0)
                
            self.draw_text(num, vector + Vector(-4 - len(num) * (size / 5), 2), 
                           color.brightened(), 50, True)
            self.draw_lozengy(color, color.dimmed(), vector + 
                              Vector(len(num) * (size / 4 - 1) + 22, 0) +
                              self.origin(), size / 2)
            self.draw_text(type(int(num)).numerus(), vector + Vector(-14, 23), 
                           color.brightened(), 76, True)
            
            if nr == 0:
                vector -= Vector(5, 0)
            
            if nr < 9:
                self.draw_line(Color(0, 0, 0),
                               vector + Vector(61, -5),
                               vector + Vector(61, 41))
                
            nr += 1
        
class MapView(Component):
    def __init__(self, area, city, tile_size, border_color, on_select):
        super(MapView, self).__init__(area)
        
        self._city           = city
        self._tile_size      = tile_size
        self._border_color   = border_color
        self._textures       = {}
        self._display_vector = None
        self._on_select      = on_select
        self._selection      = None
        self._callback       = None
        self._building       = None
        self._tiles_map      = None
        self._tiles_map_nr   = 0
        self._rendered       = set()
        
    def callback(self):
        return self._callback
    
    def building(self):
        return self._building
    
    def select(self, thing):
        self._selection = thing
        self._on_select(thing)
    
    def selection(self):
        if self._selection.__class__ is ConstructionSite:
            successor = self._selection.successor()
            
            if successor is not None:
                self.select(successor)
                
        return self._selection
    
    def is_selected(self, thing):
        return (thing == self.selection())
    
    def border_color_of(self, thing):
        if self.is_selected(thing):
            return Color(0, 0, 0)
        
        return thing.color().dimmed()
    
    def log(self, string, color = None):
        self.city().log(string, color)
        
    def city(self):
        return self._city
    
    def city_map(self):
        return self.city().map()
    
    def tile_size(self):
        return self._tile_size
    
    def tile_vector(self, size = 1):
        return Vector(self.tile_size() * size, self.tile_size() * size)
    
    def tile_origin(self, tile):
        return self.display_vector() + Vector(tile.x() * self.tile_size(),
                                              tile.y() * self.tile_size())
        
    def border_color(self):
        return self._border_color
    
    def smaller_area(self):
        return (self.display_area() < self.area())
    
    def texture(self, color, size = 1, growth = 0):
        if not self._textures.has_key(color):
            self._textures[color] = Texture(self.tile_vector(size) +
                                            Vector(growth, growth), color)
            
        return self._textures[color]
    
    def display_vector(self):
        if self._display_vector is None:
            x  = min(self.area().extent().x(),
                     self.city_map().extent().x() * self.tile_size())
            y  = min(self.area().extent().y(),
                     self.city_map().extent().y() * self.tile_size())
            dx = self.area().extent().x() - x
            dy = self.area().extent().y() - y
        
            self._display_vector = Vector(dx / 2.0, dy / 2.0)
        
        return self._display_vector
    
    def display_area(self):
        vector = self.display_vector()
        
        return self.area().cropped(vector.x(), vector.y())
    
    def tile_area(self, tile, size = 1):
        return Area(self.tile_origin(tile) + self.origin(),
                    self.tile_vector(size))
    
    def tile_at(self, vector):
        tile_vector = (vector - self.display_vector()) / self.tile_size()
        x, y        = tile_vector.int_tuple()
        tile        = self.city_map().tile(x, y)
        
        if not tile.exists():
            return None
        
        return tile

    def tiles_map(self):
        if self._tiles_map is None or self._tiles_map_nr < self.city_map().changes():
            self._tiles_map_nr = self.city_map().changes()
            self._tiles_map    = Surface((self.extent().x(), self.extent().y()))
            self._rendered     = set()
            
            for tile in self.city_map():
                self._tiles_map.blit(self.texture(tile.terrain().color()),
                                     self.tile_origin(tile).int_tuple())
                
        return self._tiles_map
        
    def mouse_tile(self):
        vector = self.mouse()
        
        if vector is None:
            return None
        
        return self.tile_at(vector)
                
    def render_goods(self, good, vector):
        number = good.number()
        
        if number <= 0:
            return

        if good.is_full():
            x = 1
        else:
            x = 0
            
        self.draw_lozengy(good.color(), good.color().dimmed(), vector,
                          int(min(math.log(number, 2) + 2 + x, 
                                  self.tile_size() / 2 - 2)))
    
    def render_tiles(self):
        shift = self.tile_vector() * Vector(0.5, 0.5)
        
        self.blit(self.tiles_map(), Vector(0, 0))
        
        for tile in self.city_map():
            goods = tile.goods()
            
            if self.is_selected(tile):
                self.draw_rectangle(SurfaceArea(self.tile_origin(tile) + self.origin(),
                                                self.tile_vector(),
                                                self.surface()), Color(0, 0, 0))
            
            if goods is not None:
                self.render_goods(goods, self.tile_origin(tile) + shift)
    
    def blit_buildings(self):
        x = y = 0
        for building in self.city().buildings():
            if building in self._rendered:
                continue
            
            if building.__class__ is not ConstructionSite:
                self._rendered.add(building)
            
            area = building.area()
            
            if area is None:
                continue
            
            tile = area.origin()
            
            if building.stands_free():
                self.tiles_map().blit(self.texture(building.color(), building.size(), -2),
                                      (self.tile_origin(tile) + Vector(1, 1)).int_tuple())
            else:
                self.tiles_map().blit(self.texture(building.color(), building.size()),
                                      self.tile_origin(tile).int_tuple())
        
    def render_buildings(self):
        shift = self.tile_vector() * Vector(0.5, 0.5)

        for building in self.city().buildings():
            area = building.area()
            
            if area is None:
                continue
            
            tile = area.origin()
                
            if building.office() is not None:
                self.draw_rectangle(SurfaceArea(self.tile_origin(building.office()) +
                                                self.origin() + Vector(1, 1),
                                                self.tile_vector() - Vector(2, 2), self.surface()),
                                    building.color().dimmed())
        
        sel = self.selection()
        
        if sel is None:
            return
        
        if issubclass(sel.__class__, Building):
            area = sel.area()
            
            if area is None:
                return
            
            tile = area.origin()
            self.draw_rectangle(SurfaceArea(self.tile_origin(tile) + self.origin(),
                                            self.tile_vector() * Vector(sel.size(), sel.size()),
                                            self.surface()), Color(0, 0, 0))
                
                
    def render_paths(self):
        shift = self.tile_vector() * Vector(0.5, 0.5)

        for person in self.city().persons():
            if not self.is_selected(person):
                continue
            
            action   = person.action()
            plan     = person.plan()
            tile     = person.start_tile()
            last     = self.tile_origin(tile)
            path     = []
            color    = Color(0, 0, 0)
            dimmed   = color.dimmed()
            
            if action is not None:
                last += self.tile_vector() * action.done_vector()
                path += [self.tile_vector() * -action.remaining_vector()]
                
            if plan is not None:
                color  = plan.color()
                dimmed = color.dimmed()
                path  += map(lambda x: self.tile_vector() * x.vector(),
                             plan.path())
            
            for element in path:
                current = last + element
                
                self.draw_line(dimmed, last + shift, current + shift)
                self.draw_circle(color, dimmed, last + shift, 3)
                self.draw_circle(color, dimmed, current + shift, 3)
                
                last = current
                
    def render_people(self):
        shift = self.tile_vector() * Vector(0.5, 0.5)

        for person in self.city().persons():
            tile     = person.tile()
            goods    = person.backpack()
            position = self.tile_vector() * person.displacement()
            
            self.draw_circle(person.color(), self.border_color_of(person),
                             self.tile_origin(tile) + shift + position,
                             self.tile_size() / 2)
            
            if goods is not None:
                self.render_goods(goods,
                                  self.tile_origin(tile) + shift  + position)

    def render_construction(self):
        if self.callback() is None:
            return
        
        if not self.has_mouse():
            return
        
        building = self.building()
        tile     = self.mouse_tile()
        
        if not tile:
            return
        
        area = self.tile_area(tile, building.size())
        
        self.draw_rectangle(area, building.color())
        self.draw_rectangle(area.cropped(1), building.color())
        
        if not self.city().can_build(tile, building):
            cross = area.cropped(2).moved(-self.origin().x(),
                                          -self.origin().y())
            
            self.draw_line(Color(0, 0, 0), cross.origin(), cross.far_point() + Vector(0, 1))
            self.draw_line(Color(0, 0, 0), Vector(cross.origin().x(),
                                                  cross.far_point().y() + 1),
                           Vector(cross.far_point().x(),
                                  cross.origin().y()))

    def person_at_pixel(self, vector):
        shift = self.tile_vector() * Vector(0.5, 0.5)
        found = None
        
        for person in self.city().persons():
            tile     = person.tile()
            position = self.tile_vector() * person.displacement()
            pvector  = self.tile_origin(tile) + shift + position
            
            if (vector - pvector).length() <= self.tile_size() / 2:
                found = person
        
        return found
        
    def do_right_click(self, vector):
        if self.callback() is not None:
            self.callback()(None, self.building())
        else:
            self.select(None)        
        
    def do_left_click(self, vector):
        tile = self.tile_at(vector)
        
        if not tile or not tile.exists():
            return
        
        if self.callback() is not None:
            self.callback()(tile, self.building())
            return
        
        person   = self.person_at_pixel(vector)
        building = tile.building()
        
        if person is not None:
            self.select(person)
        elif building is not None:
            self.select(building)
        else:
            self.select(tile)
            
    def do_middle_click(self, vector):
        tile = self.tile_at(vector)
        
        if not tile or not tile.exists():
            return

        building = self.city().building_on(tile)
        
        if building is not None:
            self.select(building)
        else:
            self.select(tile)
            
    def set_build_callback(self, callback, building):
        self._callback = callback
        self._building = building
        
    def render(self):
        if self.smaller_area():
            self.draw_rectangle(self.display_area().cropped(-1),
                                self.border_color())
        
        self.blit_buildings()
        self.render_tiles()
        self.render_buildings()
        self.render_paths()
        self.render_people()
        self.render_construction()
        