'''Component base classes.'''


from pygame import *

from concepts.color import *
from concepts.geometry import *

from texture import *

import pygame.draw
import pygame.font


class SurfaceArea(Area):
    def __init__(self, origin, extent, surface):
        super(SurfaceArea, self).__init__(origin, extent)
        
        self._surface = surface
        
    def surface(self):
        return self._surface
    
    def shrunken(self, dx, dy = None):
        area = super(SurfaceArea, self).shrunken(dx, dy)
        
        return SurfaceArea(area.origin(), area.extent(), self.surface())
    
    def grown(self, dx, dy = None):
        area = super(SurfaceArea, self).grown(dx, dy)
        
        return SurfaceArea(area.origin(), area.extent(), self.surface())
    
    def moved(self, dx, dy = None):
        area = super(SurfaceArea, self).moved(dx, dy)
        
        return SurfaceArea(area.origin(), area.extent(), self.surface())
    
    def with_width(self, width):
        return SurfaceArea(self.origin(), Vector(width, self.extent().y()),
                           self.surface())
    
    def with_height(self, height):
        return SurfaceArea(self.origin(), Vector(self.extent().x(), height),
                           self.surface())
    
    def __eq__(self, other):
        return (other is not None and
                self.__class__ == other.__class__ and
                self.origin() == other.origin() and
                self.extent() == other.extent() and
                self.surface() == other.surface())

    def __hash__(self):
        return (self.origin().__hash__() + self.extent().__hash__() +
                self.surface().__hash__())

    def __repr__(self):
        return 'SurfaceArea(%s, %s, %s)' % (repr(self.origin()),
                                            repr(self.extent()),
                                            repr(self.surface()))
    
    def __str__(self):
        return '[%s-%s]' % (str(self.origin()), str(self.extent()))


class Container(object):
    def __init__(self):
        self._components = []
    
    def is_container(self):
        return True
    
    def surface(self):
        return self.area().surface()
    
    def components(self):
        return list(self._components)
        
    def add(self, component):
        self._components.append(component)
        
    def remove(self, component):
        self._components.remove(component)

    def remove_all(self):
        self._components = []
        
    def render(self):
        for component in self.components():
            component.render()

    
class Component(object):
    _font       = None
    _font_size  = None
    _small_font = None
    _logo       = None
    
    def __init__(self, area):
        self._area   = area
        self._exists = True
        self._mouse  = None
        
    @classmethod
    def font(cls):
        return cls._font
    
    @classmethod
    def small_font(cls):
        return cls._small_font
    
    @classmethod
    def font_size(cls):
        return cls._font_size
    
    @classmethod
    def small_font_size(cls):
        return int(cls._font_size * 0.8)
    
    @classmethod
    def set_font(cls, file_name, size):
        pygame.font.init()

        cls._font_size  = size
        cls._font       = pygame.font.Font(file_name, size)
        cls._small_font = pygame.font.Font(file_name, cls.small_font_size())
    
    @classmethod
    def logo(cls):
        return cls._logo
    
    @classmethod
    def set_logo(cls, file_name):
        cls._logo = file_name
    
    def mouse(self):
        return self._mouse
    
    def has_mouse(self):
        return (self._mouse is not None)
    
    def is_container(self):
        return False
    
    def exists(self):
        return self._exists
    
    def dispose(self):
        self._exists = False
        self.render  = lambda: None
        
    def area(self):
        return self._area
    
    def origin(self):
        return self.area().origin()
    
    def extent(self):
        return self.area().extent()
    
    def surface(self):
        return self.area().surface()
    
    def translate(self, vector):
        return self.origin() + vector
    
    def normalize(self, vector):
        return self.origin() + vector * self.extent()
    
    def rect(self, area):
        return Rect(area.origin().x(), area.origin().y(),
                    area.extent().x(), area.extent().y())
    
    def set_at(self, vector, color):
        self.surface().set_at(self.translate(vector).tuple(), color.tuple())
    
    def blit(self, surface, vector = Vector(0, 0)):
        self.surface().blit(surface, self.translate(vector).tuple())
    
    def draw_text(self, line, vector, color, max_length = None, center = False,
                  small = False):
        
        font = self.font()
        
        if small:
            font = self.small_font()
            
        if max_length is not None:
            width, height = font.size(line)
        
            while width > max_length:
                line = line[:-4] + '...'
                width, height = font.size(line)
            
        surface = font.render(line, True, color.tuple())
        
        if center is None:
            vector += Vector(max_length - width, 0)
        elif center:
            vector += Vector((max_length - width) / 2, 0)
            
        self.blit(surface, vector)
        
    def draw_bar(self, position, origin, extent, color, border_color):
        area  = SurfaceArea(self.translate(origin), extent, self.surface())
        width = int(position * (extent.x() - 4))
        
        self.draw_rectangle(area, border_color)
        self.draw_rectangle(area.cropped(1), Color(0, 0, 0))
        
        if width > 0:
            self.draw_filled(area.cropped(2).with_width(width), color)
        
    def draw_rectangle(self, area, color):
        pygame.draw.rect(self.surface(), color.tuple(), self.rect(area), 1)
        
    def draw_filled(self, area, color):
        pygame.draw.rect(self.surface(), color.tuple(), self.rect(area))
    
    def draw_circle(self, fill_color, color, center, radius, line_size = 1):
        pygame.draw.circle(self.surface(), fill_color.tuple(),
                           self.translate(center).int_tuple(), radius)
        pygame.draw.circle(self.surface(), color.tuple(),
                           self.translate(center).int_tuple(), 
                           radius + line_size - 1, line_size)

    def draw_lozengy(self, fill_color, color, center, radius):
        points = map(lambda x: self.translate(x).int_tuple(),
                     [center - Vector(radius, 0),
                      center + Vector(0, radius),
                      center + Vector(radius, 0),
                      center - Vector(0, radius)])

        pygame.draw.polygon(self.surface(), fill_color.tuple(), points)
        pygame.draw.polygon(self.surface(), color.tuple(), points, 1)
#        
    def draw_line(self, color, start, end):
        pygame.draw.aaline(self.surface(), color.tuple(),
                           self.translate(start).int_tuple(),
                           self.translate(end).int_tuple())
    
    def do_left_click(self, vector):
        pass
    
    def do_right_click(self, vector):
        pass
    
    def do_middle_click(self, vector):
        pass
    
    def do_mouse(self, vector):
        self._mouse = vector
    
    def render(self):
        pass


class TextureArea(Component):
    def __init__(self, area, texture):
        super(TextureArea, self).__init__(area)
        
        self._texture      = texture
        self._full_texture = self.texture().sized(self.extent())
        
    def texture(self):
        return self._texture
    
    def full_texture(self):
        return self._full_texture
    
    def render(self):
        self.blit(self.full_texture())
    

class Frame(Component):
    def __init__(self, area, color, is_inside = True, is_3d = False):
        super(Frame, self).__init__(area)
        
        self._color     = color
        self._is_inside = is_inside
        self._is_3d     = is_3d
        
    def color(self):
        return self._color
    
    def is_inside(self):
        return self._is_inside
    
    def set_inside(self, is_inside):
        self._is_inside = is_inside
        
    def is_3d(self):
        return self._is_3d
    
    def render_line_horizontal(self, y, delta, color = None):
        self.draw_line(color, Vector(delta, y),
                       Vector(self.extent().x() - delta, y))
            
    def render_line_vertical(self, x, delta, color = None):
        self.draw_line(color, Vector(x, delta),
                       Vector(x, self.extent().y() - delta))
    
    def render(self):
        x = self.extent().x() - 1
        y = self.extent().y() - 1

        if self.is_inside():
            first_color  = self.color()
            second_color = Color(0, 0, 0)
        else:
            first_color  = Color(0, 0, 0)
            second_color = self.color()

        if self.is_3d():
            if self.is_inside():                
                first_color_a  = first_color.dimmed()
                second_color_a = second_color.dimmed()
                first_color_b  = first_color
                second_color_b = second_color
            else:
                first_color_a  = first_color
                second_color_a = second_color
                first_color_b  = first_color.dimmed()
                second_color_b = second_color.dimmed()
        else:
            first_color_a  = first_color
            second_color_a = second_color
            first_color_b  = first_color
            second_color_b = second_color
            
        self.render_line_horizontal(0,     0, first_color_a)
        self.render_line_horizontal(1,     1, second_color_a)
        self.render_line_horizontal(y - 1, 1, second_color_a)
        self.render_line_horizontal(y,     0, first_color_a)
                
        self.render_line_vertical(0,     0, first_color_b)
        self.render_line_vertical(1,     1, second_color_b)
        self.render_line_vertical(x - 1, 1, second_color_b)
        self.render_line_vertical(x,     0, first_color_b)
    

class Renderer(Component):
    def __init__(self, area, renderer, on_click = lambda x: None):
        super(Renderer, self).__init__(area)
        
        self._renderer = renderer
        self._on_click = on_click
        
    def renderer(self):
        return self._renderer
    
    def on_click(self):
        return self._on_click
    
    def render(self):
        self.renderer()(self)

    def do_left_click(self, vector):
        self.on_click()(vector)


class Panel(Container):
    def __init__(self, area, frame_color, texture_color, is_inside = True):
        super(Panel, self).__init__()
    
        self._area = area
        
        self.add(TextureArea(area, Texture(Vector(20, 20), texture_color)))
        self.add(Frame(area, frame_color, is_inside))
        
    def area(self):
        return self._area

    def client_area(self):
        return self.area().cropped(2)


class Button(Container):
    def __init__(self, area, frame_color, texture_color):
        super(Button, self).__init__()
        
        self._area  = area
        self._state = False
        self._frame = Frame(area, frame_color, False, True)
        
        self.add(TextureArea(area, Texture(Vector(20, 20), texture_color)))
        self.add(self.frame())
        self.add(Renderer(self.client_area(), self.render_contents,
                          self.do_click))
    
    def area(self):
        return self._area

    def origin(self):
        return self.area().origin()

    def extent(self):
        return self.area().extent()

    def client_area(self):
        return self.area().cropped(2)
    
    def state(self):
        return self._state
    
    def set_state(self, state):
        self._state = state
        self.frame().set_inside(state)
        
    def frame(self):
        return self._frame
    
    def render_contents(self, renderer):
        pass
        
    def do_click(self, vector):
        pass
    
    def do_left_click(self, vector):
        self.do_click(vector)
    
    def do_right_click(self, vector):
        pass
    
    def do_middle_click(self, vector):
        pass
